# future
from __future__ import annotations

# stdlib
from collections import OrderedDict
from collections.abc import Callable
from datetime import MINYEAR
from datetime import datetime
from functools import partial
import hashlib
import json
import logging
import os
from pathlib import Path
import subprocess  # nosec
import sys
from time import sleep
import traceback
from typing import Any
from typing import cast

# third party
from nacl.signing import SigningKey
from result import Err
from result import Result

# relative
from .. import __version__
from ..abstract_server import AbstractServer
from ..abstract_server import ServerSideType
from ..abstract_server import ServerType
from ..client.api import SignedSyftAPICall
from ..client.api import SyftAPI
from ..client.api import SyftAPICall
from ..client.api import SyftAPIData
from ..client.api import debox_signed_syftapicall_response
from ..client.client import SyftClient
from ..exceptions.exception import PySyftException
from ..protocol.data_protocol import PROTOCOL_TYPE
from ..protocol.data_protocol import get_data_protocol
from ..service.action.action_object import Action
from ..service.action.action_object import ActionObject
from ..service.action.action_store import ActionStore
from ..service.action.action_store import DictActionStore
from ..service.action.action_store import MongoActionStore
from ..service.action.action_store import SQLiteActionStore
from ..service.blob_storage.service import BlobStorageService
from ..service.code.user_code_service import UserCodeService
from ..service.code.user_code_stash import UserCodeStash
from ..service.context import AuthedServiceContext
from ..service.context import ServerServiceContext
from ..service.context import UnauthedServiceContext
from ..service.context import UserLoginCredentials
from ..service.job.job_stash import Job
from ..service.job.job_stash import JobStash
from ..service.job.job_stash import JobStatus
from ..service.job.job_stash import JobType
from ..service.metadata.server_metadata import ServerMetadata
from ..service.network.network_service import NetworkService
from ..service.network.utils import PeerHealthCheckTask
from ..service.notifier.notifier_service import NotifierService
from ..service.queue.base_queue import AbstractMessageHandler
from ..service.queue.base_queue import QueueConsumer
from ..service.queue.base_queue import QueueProducer
from ..service.queue.queue import APICallMessageHandler
from ..service.queue.queue import QueueManager
from ..service.queue.queue_stash import APIEndpointQueueItem
from ..service.queue.queue_stash import ActionQueueItem
from ..service.queue.queue_stash import QueueItem
from ..service.queue.queue_stash import QueueStash
from ..service.queue.zmq_queue import QueueConfig
from ..service.queue.zmq_queue import ZMQClientConfig
from ..service.queue.zmq_queue import ZMQQueueConfig
from ..service.response import SyftError
from ..service.service import AbstractService
from ..service.service import ServiceConfigRegistry
from ..service.service import UserServiceConfigRegistry
from ..service.settings.settings import ServerSettings
from ..service.settings.settings import ServerSettingsUpdate
from ..service.settings.settings_stash import SettingsStash
from ..service.user.user import User
from ..service.user.user import UserCreate
from ..service.user.user import UserView
from ..service.user.user_roles import ServiceRole
from ..service.user.user_service import UserService
from ..service.user.user_stash import UserStash
from ..service.worker.utils import DEFAULT_WORKER_IMAGE_TAG
from ..service.worker.utils import DEFAULT_WORKER_POOL_NAME
from ..service.worker.utils import create_default_image
from ..service.worker.worker_image_service import SyftWorkerImageService
from ..service.worker.worker_pool import WorkerPool
from ..service.worker.worker_pool_service import SyftWorkerPoolService
from ..service.worker.worker_pool_stash import SyftWorkerPoolStash
from ..service.worker.worker_stash import WorkerStash
from ..store.blob_storage import BlobStorageConfig
from ..store.blob_storage.on_disk import OnDiskBlobStorageClientConfig
from ..store.blob_storage.on_disk import OnDiskBlobStorageConfig
from ..store.blob_storage.seaweedfs import SeaweedFSBlobDeposit
from ..store.dict_document_store import DictStoreConfig
from ..store.document_store import StoreConfig
from ..store.linked_obj import LinkedObject
from ..store.mongo_document_store import MongoStoreConfig
from ..store.sqlite_document_store import SQLiteStoreClientConfig
from ..store.sqlite_document_store import SQLiteStoreConfig
from ..types.datetime import DATETIME_FORMAT
from ..types.syft_metaclass import Empty
from ..types.syft_object import Context
from ..types.syft_object import PartialSyftObject
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject
from ..types.uid import UID
from ..util.experimental_flags import flags
from ..util.telemetry import instrument
from ..util.util import get_dev_mode
from ..util.util import get_env
from ..util.util import get_queue_address
from ..util.util import random_name
from ..util.util import str_to_bool
from ..util.util import thread_ident
from .credentials import SyftSigningKey
from .credentials import SyftVerifyKey
from .service_registry import ServiceRegistry
from .utils import get_named_server_uid
from .utils import get_temp_dir_for_server
from .utils import remove_temp_dir_for_server
from .worker_settings import WorkerSettings

logger = logging.getLogger(__name__)

# if user code needs to be serded and its not available we can call this to refresh
# the code for a specific server UID and thread
CODE_RELOADER: dict[int, Callable] = {}


SERVER_PRIVATE_KEY = "SERVER_PRIVATE_KEY"
SERVER_UID = "SERVER_UID"
SERVER_TYPE = "SERVER_TYPE"
SERVER_NAME = "SERVER_NAME"
SERVER_SIDE_TYPE = "SERVER_SIDE_TYPE"

DEFAULT_ROOT_EMAIL = "DEFAULT_ROOT_EMAIL"
DEFAULT_ROOT_USERNAME = "DEFAULT_ROOT_USERNAME"
DEFAULT_ROOT_PASSWORD = "DEFAULT_ROOT_PASSWORD"  # nosec


def get_private_key_env() -> str | None:
    return get_env(SERVER_PRIVATE_KEY)


def get_server_type() -> str | None:
    return get_env(SERVER_TYPE, "datasite")


def get_server_name() -> str | None:
    return get_env(SERVER_NAME, None)


def get_server_side_type() -> str | None:
    return get_env(SERVER_SIDE_TYPE, "high")


def get_server_uid_env() -> str | None:
    return get_env(SERVER_UID)


def get_default_root_email() -> str | None:
    return get_env(DEFAULT_ROOT_EMAIL, "info@openmined.org")


def get_default_root_username() -> str | None:
    return get_env(DEFAULT_ROOT_USERNAME, "Jane Doe")


def get_default_root_password() -> str | None:
    return get_env(DEFAULT_ROOT_PASSWORD, "changethis")  # nosec


def get_enable_warnings() -> bool:
    return str_to_bool(get_env("ENABLE_WARNINGS", "False"))


def get_container_host() -> str | None:
    return get_env("CONTAINER_HOST")


def get_default_worker_image() -> str | None:
    return get_env("DEFAULT_WORKER_POOL_IMAGE")


def get_default_worker_pool_name() -> str | None:
    return get_env("DEFAULT_WORKER_POOL_NAME", DEFAULT_WORKER_POOL_NAME)


def get_default_bucket_name() -> str:
    env = get_env("DEFAULT_BUCKET_NAME")
    server_id = get_server_uid_env() or "syft-bucket"
    return env or server_id or "syft-bucket"


def get_default_worker_pool_count(server: Server) -> int:
    return int(
        get_env(
            "DEFAULT_WORKER_POOL_COUNT", server.queue_config.client_config.n_consumers
        )
    )


def get_default_worker_pool_pod_annotations() -> dict[str, str] | None:
    annotations = get_env("DEFAULT_WORKER_POOL_POD_ANNOTATIONS", "null")
    return json.loads(annotations)


def get_default_worker_pool_pod_labels() -> dict[str, str] | None:
    labels = get_env("DEFAULT_WORKER_POOL_POD_LABELS", "null")
    return json.loads(labels)


def in_kubernetes() -> bool:
    return get_container_host() == "k8s"


def get_venv_packages() -> str:
    try:
        # subprocess call is safe because it uses a fully qualified path and fixed arguments
        result = subprocess.run(
            [sys.executable, "-m", "pip", "list", "--format=freeze"],  # nosec
            capture_output=True,
            check=True,
            text=True,
        )
        return result.stdout
    except subprocess.CalledProcessError as e:
        return f"An error occurred: {e.stderr}"


def get_syft_worker() -> bool:
    return str_to_bool(get_env("SYFT_WORKER", "false"))


def get_k8s_pod_name() -> str | None:
    return get_env("K8S_POD_NAME")


def get_syft_worker_uid() -> str | None:
    is_worker = get_syft_worker()
    pod_name = get_k8s_pod_name()
    uid = get_env("SYFT_WORKER_UID")
    # if uid is empty is a K8S worker, generate a uid from the pod name
    if (not uid) and is_worker and pod_name:
        uid = str(UID.with_seed(pod_name))
    return uid


signing_key_env = get_private_key_env()
server_uid_env = get_server_uid_env()

default_root_email = get_default_root_email()
default_root_username = get_default_root_username()
default_root_password = get_default_root_password()


class AuthServerContextRegistry:
    __server_context_registry__: dict[str, ServerServiceContext] = OrderedDict()

    @classmethod
    def set_server_context(
        cls,
        server_uid: UID | str,
        context: ServerServiceContext,
        user_verify_key: SyftVerifyKey | str,
    ) -> None:
        if isinstance(server_uid, str):
            server_uid = UID.from_string(server_uid)

        if isinstance(user_verify_key, str):
            user_verify_key = SyftVerifyKey.from_string(user_verify_key)

        key = cls._get_key(server_uid=server_uid, user_verify_key=user_verify_key)

        cls.__server_context_registry__[key] = context

    @staticmethod
    def _get_key(server_uid: UID, user_verify_key: SyftVerifyKey) -> str:
        return "-".join(str(x) for x in (server_uid, user_verify_key))

    @classmethod
    def auth_context_for_user(
        cls,
        server_uid: UID,
        user_verify_key: SyftVerifyKey,
    ) -> AuthedServiceContext | None:
        key = cls._get_key(server_uid=server_uid, user_verify_key=user_verify_key)
        return cls.__server_context_registry__.get(key)


@instrument
class Server(AbstractServer):
    signing_key: SyftSigningKey | None
    required_signed_calls: bool = True
    packages: str

    def __init__(
        self,
        *,  # Trasterisk
        name: str | None = None,
        id: UID | None = None,
        signing_key: SyftSigningKey | SigningKey | None = None,
        action_store_config: StoreConfig | None = None,
        document_store_config: StoreConfig | None = None,
        root_email: str | None = default_root_email,
        root_username: str | None = default_root_username,
        root_password: str | None = default_root_password,
        processes: int = 0,
        is_subprocess: bool = False,
        server_type: str | ServerType = ServerType.DATASITE,
        local_db: bool = False,
        reset: bool = False,
        blob_storage_config: BlobStorageConfig | None = None,
        queue_config: QueueConfig | None = None,
        queue_port: int | None = None,
        n_consumers: int = 0,
        create_producer: bool = False,
        thread_workers: bool = False,
        server_side_type: str | ServerSideType = ServerSideType.HIGH_SIDE,
        enable_warnings: bool = False,
        dev_mode: bool = False,
        migrate: bool = False,
        in_memory_workers: bool = True,
        smtp_username: str | None = None,
        smtp_password: str | None = None,
        email_sender: str | None = None,
        smtp_port: int | None = None,
        smtp_host: str | None = None,
        association_request_auto_approval: bool = False,
        background_tasks: bool = False,
    ):
        # 🟡 TODO 22: change our ENV variable format and default init args to make this
        # less horrible or add some convenience functions
        self.dev_mode = dev_mode or get_dev_mode()
        self.id = UID.from_string(server_uid_env) if server_uid_env else (id or UID())
        self.packages = ""
        self.processes = processes
        self.is_subprocess = is_subprocess
        self.name = name or random_name()
        self.enable_warnings = enable_warnings
        self.in_memory_workers = in_memory_workers
        self.server_type = ServerType(server_type)
        self.server_side_type = ServerSideType(server_side_type)
        self.client_cache: dict = {}
        self.peer_client_cache: dict = {}

        if isinstance(server_type, str):
            server_type = ServerType(server_type)
        self.server_type = server_type

        if isinstance(server_side_type, str):
            server_side_type = ServerSideType(server_side_type)
        self.server_side_type = server_side_type

        skey = None
        if signing_key_env:
            skey = SyftSigningKey.from_string(signing_key_env)
        elif isinstance(signing_key, SigningKey):
            skey = SyftSigningKey(signing_key=signing_key)
        else:
            skey = signing_key
        self.signing_key = skey or SyftSigningKey.generate()

        self.association_request_auto_approval = association_request_auto_approval

        self.queue_config = self.create_queue_config(
            n_consumers=n_consumers,
            create_producer=create_producer,
            thread_workers=thread_workers,
            queue_port=queue_port,
            queue_config=queue_config,
        )

        # must call before initializing stores
        if reset:
            self.remove_temp_dir()

        use_sqlite = local_db or (processes > 0 and not is_subprocess)
        document_store_config = document_store_config or self.get_default_store(
            use_sqlite=use_sqlite,
            store_type="Document Store",
        )
        action_store_config = action_store_config or self.get_default_store(
            use_sqlite=use_sqlite,
            store_type="Action Store",
        )
        self.init_stores(
            action_store_config=action_store_config,
            document_store_config=document_store_config,
        )

        # construct services only after init stores
        self.services: ServiceRegistry = ServiceRegistry.for_server(self)

        create_admin_new(  # nosec B106
            name=root_username,
            email=root_email,
            password=root_password,
            server=self,
        )

        NotifierService.init_notifier(
            server=self,
            email_password=smtp_password,
            email_username=smtp_username,
            email_sender=email_sender,
            smtp_port=smtp_port,
            smtp_host=smtp_host,
        )

        self.post_init()

        if migrate:
            self.find_and_migrate_data()
        else:
            self.find_and_migrate_data([ServerSettings])

        self.create_initial_settings(admin_email=root_email)

        self.init_blob_storage(config=blob_storage_config)

        # Migrate data before any operation on db

        # first migrate, for backwards compatibility
        self.init_queue_manager(queue_config=self.queue_config)

        context = AuthedServiceContext(
            server=self,
            credentials=self.verify_key,
            role=ServiceRole.ADMIN,
        )

        self.peer_health_manager: PeerHealthCheckTask | None = None
        if background_tasks:
            self.run_peer_health_checks(context=context)

        ServerRegistry.set_server_for(self.id, self)

    @property
    def runs_in_docker(self) -> bool:
        path = "/proc/self/cgroup"
        return (
            os.path.exists("/.dockerenv")
            or os.path.isfile(path)
            and any("docker" in line for line in open(path))
        )

    def get_default_store(self, use_sqlite: bool, store_type: str) -> StoreConfig:
        if use_sqlite:
            path = self.get_temp_dir("db")
            file_name: str = f"{self.id}.sqlite"
            if self.dev_mode:
                logger.debug(f"{store_type}'s SQLite DB path: {path/file_name}")
            return SQLiteStoreConfig(
                client_config=SQLiteStoreClientConfig(
                    filename=file_name,
                    path=path,
                )
            )
        return DictStoreConfig()

    def init_blob_storage(self, config: BlobStorageConfig | None = None) -> None:
        if config is None:
            client_config = OnDiskBlobStorageClientConfig(
                base_directory=self.get_temp_dir("blob")
            )
            config_ = OnDiskBlobStorageConfig(
                client_config=client_config,
                min_blob_size=os.getenv("MIN_SIZE_BLOB_STORAGE_MB", 16),
            )
        else:
            config_ = config
        self.blob_store_config = config_
        self.blob_storage_client = config_.client_type(config=config_.client_config)

        # relative
        from ..store.blob_storage.seaweedfs import SeaweedFSConfig

        if isinstance(config, SeaweedFSConfig) and self.signing_key:
            blob_storage_service = self.get_service(BlobStorageService)
            remote_profiles = blob_storage_service.remote_profile_stash.get_all(
                credentials=self.signing_key.verify_key, has_permission=True
            ).ok()
            for remote_profile in remote_profiles:
                self.blob_store_config.client_config.remote_profiles[
                    remote_profile.profile_name
                ] = remote_profile

        if self.dev_mode:
            if isinstance(self.blob_store_config, OnDiskBlobStorageConfig):
                logger.debug(
                    f"Using on-disk blob storage with path: "
                    f"{self.blob_store_config.client_config.base_directory}",
                )
            logger.debug(
                f"Minimum object size to be saved to the blob storage: "
                f"{self.blob_store_config.min_blob_size} (MB)."
            )

    def run_peer_health_checks(self, context: AuthedServiceContext) -> None:
        self.peer_health_manager = PeerHealthCheckTask()
        self.peer_health_manager.run(context=context)

    def stop(self) -> None:
        if self.peer_health_manager is not None:
            self.peer_health_manager.stop()

        for consumer_list in self.queue_manager.consumers.values():
            for c in consumer_list:
                c.close()
        for p in self.queue_manager.producers.values():
            p.close()

        self.queue_manager.producers.clear()
        self.queue_manager.consumers.clear()

        ServerRegistry.remove_server(self.id)

    def close(self) -> None:
        self.stop()

    def cleanup(self) -> None:
        self.stop()
        self.remove_temp_dir()

    def create_queue_config(
        self,
        n_consumers: int,
        create_producer: bool,
        thread_workers: bool,
        queue_port: int | None,
        queue_config: QueueConfig | None,
    ) -> QueueConfig:
        if queue_config:
            queue_config_ = queue_config
        elif queue_port is not None or n_consumers > 0 or create_producer:
            if not create_producer and queue_port is None:
                logger.warn("No queue port defined to bind consumers.")
            queue_config_ = ZMQQueueConfig(
                client_config=ZMQClientConfig(
                    create_producer=create_producer,
                    queue_port=queue_port,
                    n_consumers=n_consumers,
                ),
                thread_workers=thread_workers,
            )
        else:
            queue_config_ = ZMQQueueConfig()

        return queue_config_

    def init_queue_manager(self, queue_config: QueueConfig) -> None:
        MessageHandlers = [APICallMessageHandler]
        if self.is_subprocess:
            return None

        self.queue_manager = QueueManager(config=queue_config)
        for message_handler in MessageHandlers:
            queue_name = message_handler.queue_name
            # client config
            if getattr(queue_config.client_config, "create_producer", True):
                context = AuthedServiceContext(
                    server=self,
                    credentials=self.verify_key,
                    role=ServiceRole.ADMIN,
                )
                producer: QueueProducer = self.queue_manager.create_producer(
                    queue_name=queue_name,
                    queue_stash=self.queue_stash,
                    context=context,
                    worker_stash=self.worker_stash,
                )
                producer.run()
                address = producer.address
            else:
                port = queue_config.client_config.queue_port
                if port is not None:
                    address = get_queue_address(port)
                else:
                    address = None

            if address is None and queue_config.client_config.n_consumers > 0:
                raise ValueError("address unknown for consumers")

            service_name = queue_config.client_config.consumer_service

            if not service_name:
                # Create consumers for default worker pool
                create_default_worker_pool(self)
            else:
                # Create consumer for given worker pool
                syft_worker_uid = get_syft_worker_uid()
                logger.info(
                    f"Running as consumer with uid={syft_worker_uid} service={service_name}"
                )

                if syft_worker_uid:
                    self.add_consumer_for_service(
                        service_name=service_name,
                        syft_worker_id=UID(syft_worker_uid),
                        address=address,
                        message_handler=message_handler,
                    )

    def add_consumer_for_service(
        self,
        service_name: str,
        syft_worker_id: UID,
        address: str,
        message_handler: type[AbstractMessageHandler] = APICallMessageHandler,
    ) -> None:
        consumer: QueueConsumer = self.queue_manager.create_consumer(
            message_handler,
            address=address,
            service_name=service_name,
            worker_stash=self.worker_stash,
            syft_worker_id=syft_worker_id,
        )
        consumer.run()

    def remove_consumer_with_id(self, syft_worker_id: UID) -> None:
        for consumers in self.queue_manager.consumers.values():
            # Grab the list of consumers for the given queue
            consumer_to_pop = None
            for consumer_idx, consumer in enumerate(consumers):
                if consumer.syft_worker_id == syft_worker_id:
                    consumer.close()
                    consumer_to_pop = consumer_idx
                    break
            if consumer_to_pop is not None:
                consumers.pop(consumer_to_pop)

    @classmethod
    def named(
        cls: type[Server],
        *,  # Trasterisk
        name: str,
        processes: int = 0,
        reset: bool = False,
        local_db: bool = False,
        server_type: str | ServerType = ServerType.DATASITE,
        server_side_type: str | ServerSideType = ServerSideType.HIGH_SIDE,
        enable_warnings: bool = False,
        n_consumers: int = 0,
        thread_workers: bool = False,
        create_producer: bool = False,
        queue_port: int | None = None,
        dev_mode: bool = False,
        migrate: bool = False,
        in_memory_workers: bool = True,
        association_request_auto_approval: bool = False,
        background_tasks: bool = False,
    ) -> Server:
        uid = get_named_server_uid(name)
        name_hash = hashlib.sha256(name.encode("utf8")).digest()
        key = SyftSigningKey(signing_key=SigningKey(name_hash))
        blob_storage_config = None

        server_type = ServerType(server_type)
        server_side_type = ServerSideType(server_side_type)

        return cls(
            name=name,
            id=uid,
            signing_key=key,
            processes=processes,
            local_db=local_db,
            server_type=server_type,
            server_side_type=server_side_type,
            enable_warnings=enable_warnings,
            blob_storage_config=blob_storage_config,
            queue_port=queue_port,
            n_consumers=n_consumers,
            thread_workers=thread_workers,
            create_producer=create_producer,
            dev_mode=dev_mode,
            migrate=migrate,
            in_memory_workers=in_memory_workers,
            reset=reset,
            association_request_auto_approval=association_request_auto_approval,
            background_tasks=background_tasks,
        )

    def is_root(self, credentials: SyftVerifyKey) -> bool:
        return credentials == self.verify_key

    @property
    def root_client(self) -> SyftClient:
        # relative
        from ..client.client import PythonConnection

        connection = PythonConnection(server=self)
        client_type = connection.get_client_type()
        if isinstance(client_type, SyftError):
            return client_type
        root_client = client_type(connection=connection, credentials=self.signing_key)
        if root_client.api.refresh_api_callback is not None:
            root_client.api.refresh_api_callback()
        return root_client

    def _find_klasses_pending_for_migration(
        self, object_types: list[SyftObject]
    ) -> list[SyftObject]:
        context = AuthedServiceContext(
            server=self,
            credentials=self.verify_key,
            role=ServiceRole.ADMIN,
        )
        migration_state_service = self.services.migration

        klasses_to_be_migrated = []

        for object_type in object_types:
            canonical_name = object_type.__canonical_name__
            object_version = object_type.__version__

            migration_state = migration_state_service.get_state(context, canonical_name)
            if isinstance(migration_state, SyftError):
                raise Exception(
                    f"Failed to get migration state for {canonical_name}. Error: {migration_state}"
                )
            if (
                migration_state is not None
                and migration_state.current_version != migration_state.latest_version
            ):
                klasses_to_be_migrated.append(object_type)
            else:
                migration_state_service.register_migration_state(
                    context,
                    current_version=object_version,
                    canonical_name=canonical_name,
                )

        return klasses_to_be_migrated

    def find_and_migrate_data(
        self, document_store_object_types: list[type[SyftObject]] | None = None
    ) -> None:
        context = AuthedServiceContext(
            server=self,
            credentials=self.verify_key,
            role=ServiceRole.ADMIN,
        )
        migration_service = self.get_service("migrationservice")
        return migration_service.migrate_data(context, document_store_object_types)

    @property
    def guest_client(self) -> SyftClient:
        return self.get_guest_client()

    @property
    def current_protocol(self) -> str | int:
        data_protocol = get_data_protocol()
        return data_protocol.latest_version

    def get_guest_client(self, verbose: bool = True) -> SyftClient:
        # relative
        from ..client.client import PythonConnection

        connection = PythonConnection(server=self)
        if verbose and self.server_side_type:
            message: str = (
                f"Logged into <{self.name}: {self.server_side_type.value.capitalize()} "
            )
            if self.server_type:
                message += f"side {self.server_type.value.capitalize()} > as GUEST"
            logger.debug(message)

        client_type = connection.get_client_type()
        if isinstance(client_type, SyftError):
            return client_type

        guest_client = client_type(
            connection=connection, credentials=SyftSigningKey.generate()
        )
        if guest_client.api.refresh_api_callback is not None:
            guest_client.api.refresh_api_callback()
        return guest_client

    def __repr__(self) -> str:
        service_string = ""
        if not self.is_subprocess:
            services = [service.__name__ for service in self.services]
            service_string = ", ".join(sorted(services))
            service_string = f"\n\nServices:\n{service_string}"
        return f"{type(self).__name__}: {self.name} - {self.id} - {self.server_type}{service_string}"

    def post_init(self) -> None:
        context = AuthedServiceContext(
            server=self, credentials=self.verify_key, role=ServiceRole.ADMIN
        )
        AuthServerContextRegistry.set_server_context(
            server_uid=self.id, user_verify_key=self.verify_key, context=context
        )

        if "usercodeservice" in self.service_path_map:
            user_code_service = self.get_service(UserCodeService)
            user_code_service.load_user_code(context=context)

        def reload_user_code() -> None:
            user_code_service.load_user_code(context=context)

        ti = thread_ident()
        if ti is not None:
            CODE_RELOADER[ti] = reload_user_code

    def init_stores(
        self,
        document_store_config: StoreConfig,
        action_store_config: StoreConfig,
    ) -> None:
        # We add the python id of the current server in order
        # to create one connection per Server object in MongoClientCache
        # so that we avoid closing the connection from a
        # different thread through the garbage collection
        if isinstance(document_store_config, MongoStoreConfig):
            document_store_config.client_config.server_obj_python_id = id(self)

        self.document_store_config = document_store_config
        self.document_store = document_store_config.store_type(
            server_uid=self.id,
            root_verify_key=self.verify_key,
            store_config=document_store_config,
        )

        if isinstance(action_store_config, SQLiteStoreConfig):
            self.action_store: ActionStore = SQLiteActionStore(
                server_uid=self.id,
                store_config=action_store_config,
                root_verify_key=self.verify_key,
                document_store=self.document_store,
            )
        elif isinstance(action_store_config, MongoStoreConfig):
            # We add the python id of the current server in order
            # to create one connection per Server object in MongoClientCache
            # so that we avoid closing the connection from a
            # different thread through the garbage collection
            action_store_config.client_config.server_obj_python_id = id(self)

            self.action_store = MongoActionStore(
                server_uid=self.id,
                root_verify_key=self.verify_key,
                store_config=action_store_config,
                document_store=self.document_store,
            )
        else:
            self.action_store = DictActionStore(
                server_uid=self.id,
                root_verify_key=self.verify_key,
                document_store=self.document_store,
            )

        self.action_store_config = action_store_config
        self.queue_stash = QueueStash(store=self.document_store)

    @property
    def job_stash(self) -> JobStash:
        return self.get_service("jobservice").stash

    @property
    def worker_stash(self) -> WorkerStash:
        return self.get_service("workerservice").stash

    @property
    def service_path_map(self) -> dict[str, AbstractService]:
        return self.services.service_path_map

    @property
    def initialized_services(self) -> list[AbstractService]:
        return self.services.services

    def get_service_method(self, path_or_func: str | Callable) -> Callable:
        if callable(path_or_func):
            path_or_func = path_or_func.__qualname__
        return self._get_service_method_from_path(path_or_func)

    def get_service(self, path_or_func: str | Callable) -> AbstractService:
        return self.services.get_service(path_or_func)

    def _get_service_method_from_path(self, path: str) -> Callable:
        path_list = path.split(".")
        method_name = path_list.pop()
        service_obj = self.services._get_service_from_path(path=path)

        return getattr(service_obj, method_name)

    def get_temp_dir(self, dir_name: str = "") -> Path:
        """
        Get a temporary directory unique to the server.
        Provide all dbs, blob dirs, and locks using this directory.
        """
        return get_temp_dir_for_server(self.id, dir_name)

    def remove_temp_dir(self) -> None:
        """
        Remove the temporary directory for this server.
        """
        remove_temp_dir_for_server(self.id)

    def update_self(self, settings: ServerSettings) -> None:
        updateable_attrs = (
            ServerSettingsUpdate.model_fields.keys()
            - PartialSyftObject.model_fields.keys()
        )
        for attr_name in updateable_attrs:
            attr = getattr(settings, attr_name)
            if attr is not Empty:
                setattr(self, attr_name, attr)

    @property
    def settings(self) -> ServerSettings:
        settings_stash = SettingsStash(store=self.document_store)
        if self.signing_key is None:
            raise ValueError(f"{self} has no signing key")
        settings = settings_stash.get_all(self.signing_key.verify_key)
        if settings.is_err():
            raise ValueError(
                f"Cannot get server settings for '{self.name}'. Error: {settings.err()}"
            )
        if settings.is_ok() and len(settings.ok()) > 0:
            settings = settings.ok()[0]
        self.update_self(settings)
        return settings

    @property
    def metadata(self) -> ServerMetadata:
        show_warnings = self.enable_warnings
        settings_data = self.settings
        name = settings_data.name
        organization = settings_data.organization
        description = settings_data.description
        show_warnings = settings_data.show_warnings
        server_type = (
            settings_data.server_type.value if settings_data.server_type else ""
        )
        server_side_type = (
            settings_data.server_side_type.value
            if settings_data.server_side_type
            else ""
        )
        eager_execution_enabled = settings_data.eager_execution_enabled

        return ServerMetadata(
            name=name,
            id=self.id,
            verify_key=self.verify_key,
            highest_version=SYFT_OBJECT_VERSION_1,
            lowest_version=SYFT_OBJECT_VERSION_1,
            syft_version=__version__,
            description=description,
            organization=organization,
            server_type=server_type,
            server_side_type=server_side_type,
            show_warnings=show_warnings,
            eager_execution_enabled=eager_execution_enabled,
            min_size_blob_storage_mb=self.blob_store_config.min_blob_size,
        )

    @property
    def icon(self) -> str:
        return "🦾"

    @property
    def verify_key(self) -> SyftVerifyKey:
        if self.signing_key is None:
            raise ValueError(f"{self} has no signing key")
        return self.signing_key.verify_key

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False

        if self.id != other.id:
            return False

        return True

    def await_future(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> QueueItem | None | SyftError:
        # stdlib

        # relative
        from ..service.queue.queue import Status

        while True:
            result = self.queue_stash.pop_on_complete(credentials, uid)
            if not result.is_ok():
                return result.err()
            else:
                res = result.ok()
                if res.status == Status.COMPLETED:
                    return res
            sleep(0.1)

    def resolve_future(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> QueueItem | None | SyftError:
        result = self.queue_stash.pop_on_complete(credentials, uid)

        if result.is_ok():
            queue_obj = result.ok()
            queue_obj._set_obj_location_(
                server_uid=self.id,
                credentials=credentials,
            )
            return queue_obj
        return result.err()

    def forward_message(
        self, api_call: SyftAPICall | SignedSyftAPICall
    ) -> Result | QueueItem | SyftObject | SyftError | Any:
        server_uid = api_call.message.server_uid
        if "networkservice" not in self.service_path_map:
            return SyftError(
                message=(
                    "Server has no network service so we can't "
                    f"forward this message to {server_uid}"
                )
            )

        client = None

        network_service = self.get_service(NetworkService)
        peer = network_service.stash.get_by_uid(self.verify_key, server_uid)

        if peer.is_ok() and peer.ok():
            peer = peer.ok()

            # Since we have several routes to a peer
            # we need to cache the client for a given server_uid along with the route
            peer_cache_key = hash(server_uid) + hash(peer.pick_highest_priority_route())
            if peer_cache_key in self.peer_client_cache:
                client = self.peer_client_cache[peer_cache_key]
            else:
                context = AuthedServiceContext(
                    server=self, credentials=api_call.credentials
                )

                client = peer.client_with_context(context=context)
                if client.is_err():
                    return SyftError(
                        message=f"Failed to create remote client for peer: "
                        f"{peer.id}. Error: {client.err()}"
                    )
                client = client.ok()

                self.peer_client_cache[peer_cache_key] = client

        if client:
            message: SyftAPICall = api_call.message
            if message.path == "metadata":
                result = client.metadata
            elif message.path == "login":
                result = client.connection.login(**message.kwargs)
            elif message.path == "register":
                result = client.connection.register(**message.kwargs)
            elif message.path == "api":
                result = client.connection.get_api(**message.kwargs)
            else:
                signed_result = client.connection.make_call(api_call)
                result = debox_signed_syftapicall_response(signed_result=signed_result)

                # relative
                from ..store.blob_storage import BlobRetrievalByURL

                if isinstance(result, BlobRetrievalByURL | SeaweedFSBlobDeposit):
                    result.proxy_server_uid = peer.id

            return result

        return SyftError(message=(f"Server has no route to {server_uid}"))

    def get_role_for_credentials(self, credentials: SyftVerifyKey) -> ServiceRole:
        role = self.get_service("userservice").get_role_for_credentials(
            credentials=credentials
        )
        return role

    def handle_api_call(
        self,
        api_call: SyftAPICall | SignedSyftAPICall,
        job_id: UID | None = None,
        check_call_location: bool = True,
    ) -> Result[SignedSyftAPICall, Err]:
        # Get the result
        result = self.handle_api_call_with_unsigned_result(
            api_call, job_id=job_id, check_call_location=check_call_location
        )
        # Sign the result
        signed_result = SyftAPIData(data=result).sign(self.signing_key)

        return signed_result

    def handle_api_call_with_unsigned_result(
        self,
        api_call: SyftAPICall | SignedSyftAPICall,
        job_id: UID | None = None,
        check_call_location: bool = True,
    ) -> Result | QueueItem | SyftObject | SyftError:
        if self.required_signed_calls and isinstance(api_call, SyftAPICall):
            return SyftError(
                message=f"You sent a {type(api_call)}. This server requires SignedSyftAPICall."
            )
        else:
            if not api_call.is_valid:
                return SyftError(message="Your message signature is invalid")

        if api_call.message.server_uid != self.id and check_call_location:
            return self.forward_message(api_call=api_call)

        if api_call.message.path == "queue":
            return self.resolve_future(
                credentials=api_call.credentials, uid=api_call.message.kwargs["uid"]
            )

        if api_call.message.path == "metadata":
            return self.metadata

        result = None
        is_blocking = api_call.message.blocking

        if is_blocking or self.is_subprocess:
            credentials: SyftVerifyKey = api_call.credentials
            api_call = api_call.message

            role = self.get_role_for_credentials(credentials=credentials)
            context = AuthedServiceContext(
                server=self,
                credentials=credentials,
                role=role,
                job_id=job_id,
                is_blocking_api_call=is_blocking,
            )
            AuthServerContextRegistry.set_server_context(self.id, context, credentials)

            user_config_registry = UserServiceConfigRegistry.from_role(role)

            if api_call.path not in user_config_registry:
                if ServiceConfigRegistry.path_exists(api_call.path):
                    return SyftError(
                        message=f"As a `{role}`, "
                        f"you have no access to: {api_call.path}"
                    )
                else:
                    return SyftError(
                        message=f"API call not in registered services: {api_call.path}"
                    )

            _private_api_path = user_config_registry.private_path_for(api_call.path)
            method = self.get_service_method(_private_api_path)
            try:
                logger.info(f"API Call: {api_call}")
                result = method(context, *api_call.args, **api_call.kwargs)
            except PySyftException as e:
                return e.handle()
            except Exception:
                result = SyftError(
                    message=f"Exception calling {api_call.path}. {traceback.format_exc()}"
                )
        else:
            return self.add_api_call_to_queue(api_call)
        return result

    def add_api_endpoint_execution_to_queue(
        self,
        credentials: SyftVerifyKey,
        method: str,
        path: str,
        *args: Any,
        worker_pool: str | None = None,
        **kwargs: Any,
    ) -> Job | SyftError:
        job_id = UID()
        task_uid = UID()
        worker_settings = WorkerSettings.from_server(server=self)

        if worker_pool is None:
            worker_pool = self.get_default_worker_pool()
        else:
            worker_pool = self.get_worker_pool_by_name(worker_pool)

        if isinstance(worker_pool, SyftError):
            return worker_pool
        elif worker_pool is None:
            return SyftError(message="Worker pool not found")

        # Create a Worker pool reference object
        worker_pool_ref = LinkedObject.from_obj(
            worker_pool,
            service_type=SyftWorkerPoolService,
            server_uid=self.id,
        )
        queue_item = APIEndpointQueueItem(
            id=task_uid,
            method=method,
            server_uid=self.id,
            syft_client_verify_key=credentials,
            syft_server_location=self.id,
            job_id=job_id,
            worker_settings=worker_settings,
            args=args,
            kwargs={"path": path, **kwargs},
            has_execute_permissions=True,
            worker_pool=worker_pool_ref,  # set worker pool reference as part of queue item
        )

        action = Action.from_api_endpoint_execution()
        return self.add_queueitem_to_queue(
            queue_item=queue_item,
            credentials=credentials,
            action=action,
            job_type=JobType.TWINAPIJOB,
        )

    def get_worker_pool_ref_by_name(
        self, credentials: SyftVerifyKey, worker_pool_name: str | None = None
    ) -> LinkedObject | SyftError:
        # If worker pool id is not set, then use default worker pool
        # Else, get the worker pool for given uid
        if worker_pool_name is None:
            worker_pool = self.get_default_worker_pool()
        else:
            result = self.pool_stash.get_by_name(credentials, worker_pool_name)
            if result.is_err():
                return SyftError(message=f"{result.err()}")
            worker_pool = result.ok()

        # Create a Worker pool reference object
        worker_pool_ref = LinkedObject.from_obj(
            worker_pool,
            service_type=SyftWorkerPoolService,
            server_uid=self.id,
        )
        return worker_pool_ref

    def add_action_to_queue(
        self,
        action: Action,
        credentials: SyftVerifyKey,
        parent_job_id: UID | None = None,
        has_execute_permissions: bool = False,
        worker_pool_name: str | None = None,
    ) -> Job | SyftError:
        job_id = UID()
        task_uid = UID()
        worker_settings = WorkerSettings.from_server(server=self)

        # Extract worker pool id from user code
        if action.user_code_id is not None:
            result = self.user_code_stash.get_by_uid(
                credentials=credentials, uid=action.user_code_id
            )

            # If result is Ok, then user code object exists
            if result.is_ok() and result.ok() is not None:
                user_code = result.ok()
                worker_pool_name = user_code.worker_pool_name

        worker_pool_ref = self.get_worker_pool_ref_by_name(
            credentials, worker_pool_name
        )
        if isinstance(worker_pool_ref, SyftError):
            return worker_pool_ref
        queue_item = ActionQueueItem(
            id=task_uid,
            server_uid=self.id,
            syft_client_verify_key=credentials,
            syft_server_location=self.id,
            job_id=job_id,
            worker_settings=worker_settings,
            args=[],
            kwargs={"action": action},
            has_execute_permissions=has_execute_permissions,
            worker_pool=worker_pool_ref,  # set worker pool reference as part of queue item
        )
        user_id = self.get_service("UserService").get_user_id_for_credentials(
            credentials
        )

        return self.add_queueitem_to_queue(
            queue_item=queue_item,
            credentials=credentials,
            action=action,
            parent_job_id=parent_job_id,
            user_id=user_id,
        )

    def add_queueitem_to_queue(
        self,
        *,
        queue_item: QueueItem,
        credentials: SyftVerifyKey,
        action: Action | None = None,
        parent_job_id: UID | None = None,
        user_id: UID | None = None,
        job_type: JobType = JobType.JOB,
    ) -> Job | SyftError:
        log_id = UID()
        role = self.get_role_for_credentials(credentials=credentials)
        context = AuthedServiceContext(server=self, credentials=credentials, role=role)

        result_obj = ActionObject.empty()
        if action is not None:
            result_obj = ActionObject.obj_not_ready(id=action.result_id)
            result_obj.id = action.result_id
            result_obj.syft_resolved = False
            result_obj.syft_server_location = self.id
            result_obj.syft_client_verify_key = credentials

            action_service = self.get_service("actionservice")

            if not action_service.store.exists(uid=action.result_id):
                result = action_service.set_result_to_store(
                    result_action_object=result_obj,
                    context=context,
                )
                if result.is_err():
                    return result.err()

        job = Job(
            id=queue_item.job_id,
            result=result_obj,
            server_uid=self.id,
            syft_client_verify_key=credentials,
            syft_server_location=self.id,
            log_id=log_id,
            parent_job_id=parent_job_id,
            action=action,
            requested_by=user_id,
            job_type=job_type,
        )

        # 🟡 TODO 36: Needs distributed lock
        job_res = self.job_stash.set(credentials, job)
        if job_res.is_err():
            return SyftError(message=f"{job_res.err()}")
        self.queue_stash.set_placeholder(credentials, queue_item)

        log_service = self.get_service("logservice")

        result = log_service.add(context, log_id, queue_item.job_id)
        if isinstance(result, SyftError):
            return result
        return job

    def _sort_jobs(self, jobs: list[Job]) -> list[Job]:
        job_datetimes = {}
        for job in jobs:
            try:
                d = datetime.strptime(job.creation_time, DATETIME_FORMAT)
            except Exception:
                d = datetime(MINYEAR, 1, 1)
            job_datetimes[job.id] = d

        jobs.sort(
            key=lambda job: (job.status != JobStatus.COMPLETED, job_datetimes[job.id]),
            reverse=True,
        )

        return jobs

    def _get_existing_user_code_jobs(
        self, context: AuthedServiceContext, user_code_id: UID
    ) -> list[Job] | SyftError:
        job_service = self.get_service("jobservice")
        jobs = job_service.get_by_user_code_id(
            context=context, user_code_id=user_code_id
        )

        if isinstance(jobs, SyftError):
            return jobs

        return self._sort_jobs(jobs)

    def _is_usercode_call_on_owned_kwargs(
        self,
        context: AuthedServiceContext,
        api_call: SyftAPICall,
        user_code_id: UID,
    ) -> bool:
        if api_call.path != "code.call":
            return False
        user_code_service = self.get_service("usercodeservice")
        return user_code_service.is_execution_on_owned_args(
            context, user_code_id, api_call.kwargs
        )

    def add_api_call_to_queue(
        self, api_call: SyftAPICall, parent_job_id: UID | None = None
    ) -> Job | SyftError:
        unsigned_call = api_call
        if isinstance(api_call, SignedSyftAPICall):
            unsigned_call = api_call.message

        credentials = api_call.credentials
        context = AuthedServiceContext(
            server=self,
            credentials=credentials,
            role=self.get_role_for_credentials(credentials=credentials),
        )

        is_user_code = unsigned_call.path == "code.call"

        service_str, method_str = unsigned_call.path.split(".")

        action = None
        if is_user_code:
            action = Action.from_api_call(unsigned_call)
            user_code_id = action.user_code_id

            user = self.get_service(UserService).get_current_user(context)
            if isinstance(user, SyftError):
                return user
            user = cast(UserView, user)

            is_execution_on_owned_kwargs_allowed = (
                user.mock_execution_permission or context.role == ServiceRole.ADMIN
            )
            is_usercode_call_on_owned_kwargs = self._is_usercode_call_on_owned_kwargs(
                context, unsigned_call, user_code_id
            )
            # Low side does not execute jobs, unless this is a mock execution
            if (
                not is_usercode_call_on_owned_kwargs
                and self.server_side_type == ServerSideType.LOW_SIDE
            ):
                existing_jobs = self._get_existing_user_code_jobs(context, user_code_id)
                if isinstance(existing_jobs, SyftError):
                    return existing_jobs
                elif len(existing_jobs) > 0:
                    # Print warning if there are existing jobs for this user code
                    # relative
                    from ..util.util import prompt_warning_message

                    prompt_warning_message(
                        "There are existing jobs for this user code, returning the latest one"
                    )
                    return existing_jobs[-1]
                else:
                    return SyftError(
                        message="Please wait for the admin to allow the execution of this code"
                    )

            elif (
                is_usercode_call_on_owned_kwargs
                and not is_execution_on_owned_kwargs_allowed
            ):
                return SyftError(
                    message="You do not have the permissions for mock execution, please contact the admin"
                )

            return self.add_action_to_queue(
                action, api_call.credentials, parent_job_id=parent_job_id
            )

        else:
            worker_settings = WorkerSettings.from_server(server=self)
            worker_pool_ref = self.get_worker_pool_ref_by_name(credentials=credentials)
            if isinstance(worker_pool_ref, SyftError):
                return worker_pool_ref

            queue_item = QueueItem(
                id=UID(),
                server_uid=self.id,
                syft_client_verify_key=api_call.credentials,
                syft_server_location=self.id,
                job_id=UID(),
                worker_settings=worker_settings,
                service=service_str,
                method=method_str,
                args=unsigned_call.args,
                kwargs=unsigned_call.kwargs,
                worker_pool=worker_pool_ref,
            )
            return self.add_queueitem_to_queue(
                queue_item=queue_item,
                credentials=api_call.credentials,
                action=None,
                parent_job_id=parent_job_id,
            )

    @property
    def pool_stash(self) -> SyftWorkerPoolStash:
        return self.get_service(SyftWorkerPoolService).stash

    @property
    def user_code_stash(self) -> UserCodeStash:
        return self.get_service(UserCodeService).stash

    def get_default_worker_pool(self) -> WorkerPool | None | SyftError:
        result = self.pool_stash.get_by_name(
            credentials=self.verify_key,
            pool_name=self.settings.default_worker_pool,
        )
        if result.is_err():
            return SyftError(message=f"{result.err()}")
        worker_pool = result.ok()
        return worker_pool

    def get_worker_pool_by_name(self, name: str) -> WorkerPool | None | SyftError:
        result = self.pool_stash.get_by_name(
            credentials=self.verify_key, pool_name=name
        )
        if result.is_err():
            return SyftError(message=f"{result.err()}")
        worker_pool = result.ok()
        return worker_pool

    def get_api(
        self,
        for_user: SyftVerifyKey | None = None,
        communication_protocol: PROTOCOL_TYPE | None = None,
    ) -> SyftAPI:
        return SyftAPI.for_user(
            server=self,
            user_verify_key=for_user,
            communication_protocol=communication_protocol,
        )

    def get_method_with_context(
        self, function: Callable, context: ServerServiceContext
    ) -> Callable:
        method = self.get_service_method(function)
        return partial(method, context=context)

    def get_unauthed_context(
        self, login_credentials: UserLoginCredentials
    ) -> ServerServiceContext:
        return UnauthedServiceContext(server=self, login_credentials=login_credentials)

    def create_initial_settings(self, admin_email: str) -> ServerSettings | None:
        try:
            settings_stash = SettingsStash(store=self.document_store)
            if self.signing_key is None:
                logger.debug(
                    "create_initial_settings failed as there is no signing key"
                )
                return None
            settings_exists = settings_stash.get_all(self.signing_key.verify_key).ok()
            if settings_exists:
                server_settings = settings_exists[0]
                if server_settings.__version__ != ServerSettings.__version__:
                    context = Context()
                    server_settings = server_settings.migrate_to(
                        ServerSettings.__version__, context
                    )
                    res = settings_stash.delete_by_uid(
                        self.signing_key.verify_key, server_settings.id
                    )
                    if res.is_err():
                        raise Exception(res.value)
                    res = settings_stash.set(
                        self.signing_key.verify_key, server_settings
                    )
                    if res.is_err():
                        raise Exception(res.value)
                self.name = server_settings.name
                self.association_request_auto_approval = (
                    server_settings.association_request_auto_approval
                )
                return None
            else:
                # Currently we allow automatic user registration on enclaves,
                # as enclaves do not have superusers
                if self.server_type == ServerType.ENCLAVE:
                    flags.CAN_REGISTER = True
                new_settings = ServerSettings(
                    id=self.id,
                    name=self.name,
                    verify_key=self.verify_key,
                    server_type=self.server_type,
                    deployed_on=datetime.now().date().strftime("%m/%d/%Y"),
                    signup_enabled=flags.CAN_REGISTER,
                    admin_email=admin_email,
                    server_side_type=self.server_side_type.value,  # type: ignore
                    show_warnings=self.enable_warnings,
                    association_request_auto_approval=self.association_request_auto_approval,
                    default_worker_pool=get_default_worker_pool_name(),
                )
                result = settings_stash.set(
                    credentials=self.signing_key.verify_key, settings=new_settings
                )
                if result.is_ok():
                    return result.ok()
                return None
        except Exception as e:
            logger.error("create_initial_settings failed", exc_info=e)
            return None


def create_admin_new(
    name: str,
    email: str,
    password: str,
    server: AbstractServer,
) -> User | None:
    try:
        user_stash = UserStash(store=server.document_store)
        row_exists = user_stash.get_by_email(
            credentials=server.signing_key.verify_key, email=email
        ).ok()
        if row_exists:
            return None
        else:
            create_user = UserCreate(
                name=name,
                email=email,
                password=password,
                password_verify=password,
                role=ServiceRole.ADMIN,
            )
            # New User Initialization
            # 🟡 TODO: change later but for now this gives the main user super user automatically
            user = create_user.to(User)
            user.signing_key = server.signing_key
            user.verify_key = user.signing_key.verify_key
            result = user_stash.set(
                credentials=server.signing_key.verify_key,
                user=user,
                ignore_duplicates=True,
            )
            if result.is_ok():
                return result.ok()
            else:
                raise Exception(f"Could not create user: {result}")
    except Exception as e:
        logger.error("Unable to create new admin", exc_info=e)

    return None


class ServerRegistry:
    __server_registry__: dict[UID, Server] = {}

    @classmethod
    def set_server_for(
        cls,
        server_uid: UID | str,
        server: Server,
    ) -> None:
        if isinstance(server_uid, str):
            server_uid = UID.from_string(server_uid)

        cls.__server_registry__[server_uid] = server

    @classmethod
    def server_for(cls, server_uid: UID) -> Server:
        return cls.__server_registry__.get(server_uid, None)

    @classmethod
    def get_all_servers(cls) -> list[Server]:
        return list(cls.__server_registry__.values())

    @classmethod
    def remove_server(cls, server_uid: UID) -> None:
        if server_uid in cls.__server_registry__:
            del cls.__server_registry__[server_uid]


def get_default_worker_tag_by_env(dev_mode: bool = False) -> str | None:
    if in_kubernetes():
        return get_default_worker_image()
    elif dev_mode:
        return "local-dev"
    else:
        return __version__


def create_default_worker_pool(server: Server) -> SyftError | None:
    credentials = server.verify_key
    pull_image = not server.dev_mode
    image_stash = server.get_service(SyftWorkerImageService).stash
    default_pool_name = server.settings.default_worker_pool
    default_worker_pool = server.get_default_worker_pool()
    default_worker_tag = get_default_worker_tag_by_env(server.dev_mode)
    default_worker_pool_pod_annotations = get_default_worker_pool_pod_annotations()
    default_worker_pool_pod_labels = get_default_worker_pool_pod_labels()
    worker_count = get_default_worker_pool_count(server)
    context = AuthedServiceContext(
        server=server,
        credentials=credentials,
        role=ServiceRole.ADMIN,
    )

    if isinstance(default_worker_pool, SyftError):
        logger.error(
            f"Failed to get default worker pool {default_pool_name}. "
            f"Error: {default_worker_pool.message}"
        )
        return default_worker_pool

    logger.info(f"Creating default worker image with tag='{default_worker_tag}'. ")
    # Get/Create a default worker SyftWorkerImage
    default_image = create_default_image(
        credentials=credentials,
        image_stash=image_stash,
        tag=default_worker_tag,
        in_kubernetes=in_kubernetes(),
    )
    if isinstance(default_image, SyftError):
        logger.error(f"Failed to create default worker image: {default_image.message}")
        return default_image

    if not default_image.is_built:
        logger.info(f"Building default worker image with tag={default_worker_tag}. ")
        image_build_method = server.get_service_method(SyftWorkerImageService.build)
        # Build the Image for given tag
        result = image_build_method(
            context,
            image_uid=default_image.id,
            tag=DEFAULT_WORKER_IMAGE_TAG,
            pull_image=pull_image,
        )

        if isinstance(result, SyftError):
            logger.error(f"Failed to build default worker image: {result.message}")
            return None

    # Create worker pool if it doesn't exists
    logger.info(
        "Setting up worker pool"
        f"name={default_pool_name} "
        f"workers={worker_count} "
        f"image_uid={default_image.id} "
        f"in_memory={server.in_memory_workers}. "
    )
    if default_worker_pool is None:
        worker_to_add_ = worker_count
        create_pool_method = server.get_service_method(SyftWorkerPoolService.launch)
        result = create_pool_method(
            context,
            pool_name=default_pool_name,
            image_uid=default_image.id,
            num_workers=worker_count,
            pod_annotations=default_worker_pool_pod_annotations,
            pod_labels=default_worker_pool_pod_labels,
        )
    else:
        # Else add a worker to existing worker pool
        worker_to_add_ = max(default_worker_pool.max_count, worker_count) - len(
            default_worker_pool.worker_list
        )
        if worker_to_add_ > 0:
            add_worker_method = server.get_service_method(
                SyftWorkerPoolService.add_workers
            )
            result = add_worker_method(
                context=context,
                number=worker_to_add_,
                pool_name=default_pool_name,
            )
        else:
            return None

    if isinstance(result, SyftError):
        logger.info(f"Default worker pool error. {result.message}")
        return None

    for n in range(worker_to_add_):
        container_status = result[n]
        if container_status.error:
            logger.error(
                f"Failed to create container: Worker: {container_status.worker},"
                f"Error: {container_status.error}"
            )
            return None

    logger.info("Created default worker pool.")
    return None



# stdlib
import sys
import traceback
from typing import Any

# third party
from IPython.display import display
from result import Err

# relative
from ..serde.serializable import serializable
from ..types.base import SyftBaseModel
from ..util.util import sanitize_html


class SyftResponseMessage(SyftBaseModel):
    message: str
    _bool: bool = True
    require_api_update: bool = False

    def __getattr__(self, name: str) -> Any:
        if name in [
            "_bool",
            # "_repr_html_",
            # "message",
            # 'require_api_update',
            # '__bool__',
            # '__eq__',
            # '__repr__',
            # '__str__',
            # '_repr_html_class_',
            # '_repr_html_',
            "_ipython_canary_method_should_not_exist_",
            "_ipython_display_",
            "__canonical_name__",
            "__version__",
        ] or name.startswith("_repr"):
            return super().__getattr__(name)
        display(self)
        raise AttributeError(
            f"You have tried accessing `{name}` on a {type(self).__name__} with message: {self.message}"
        )

    def __bool__(self) -> bool:
        return self._bool

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SyftResponseMessage):
            return (
                self.message == other.message
                and self._bool == other._bool
                and self.require_api_update == other.require_api_update
            )
        return self._bool == other

    def __repr__(self) -> str:
        _class_name_ = type(self).__name__
        return f"{_class_name_}: {self.message}"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"

    def _repr_html_(self) -> str:
        return (
            f'<div class="{self._repr_html_class_}" style="padding:5px;">'
            f"<strong>{type(self).__name__}</strong>: "
            f'<pre class="{self._repr_html_class_}" style="display:inline; font-family:inherit;">'
            f"{sanitize_html(self.message)}</pre></div><br/>"
        )


@serializable(canonical_name="SyftError", version=1)
class SyftError(SyftResponseMessage):
    _bool: bool = False

    @property
    def _repr_html_class_(self) -> str:
        return "alert-danger"

    def to_result(self) -> Err:
        return Err(value=self.message)

    def __bool__(self) -> bool:
        return False


@serializable(canonical_name="SyftSuccess", version=1)
class SyftSuccess(SyftResponseMessage):
    @property
    def _repr_html_class_(self) -> str:
        return "alert-success"


@serializable(canonical_name="SyftNotReady", version=1)
class SyftNotReady(SyftResponseMessage):
    _bool: bool = False

    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"


@serializable(canonical_name="SyftWarning", version=1)
class SyftWarning(SyftResponseMessage):
    @property
    def _repr_html_class_(self) -> str:
        return "alert-warning"


@serializable(canonical_name="SyftInfo", version=1)
class SyftInfo(SyftResponseMessage):
    _bool: bool = False

    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"


@serializable(canonical_name="SyftException", version=1)
class SyftException(Exception):
    traceback: bool = False
    traceback_limit: int = 10

    @property
    def _repr_html_class_(self) -> str:
        return "alert-danger"

    def _repr_html_(self) -> str:
        return (
            f'<div class="{self._repr_html_class_}" style="padding:5px;">'
            + f"<strong>{type(self).__name__}</strong>: {sanitize_html(self.args)}</div><br />"
        )

    @staticmethod
    def format_traceback(etype: Any, evalue: Any, tb: Any, tb_offset: Any) -> str:
        line = "---------------------------------------------------------------------------\n"
        template = ""
        template += line
        template += f"{type(evalue).__name__}\n"
        template += line
        template += f"Exception: {evalue}\n"

        if evalue.traceback:
            template += line
            template += "Traceback:\n"
            tb_lines = "".join(traceback.format_tb(tb, evalue.traceback_limit)) + "\n"
            template += tb_lines
            template += line

        return template


def syft_exception_handler(
    shell: Any, etype: Any, evalue: Any, tb: Any, tb_offset: Any = None
) -> None:
    template = evalue.format_traceback(
        etype=etype, evalue=evalue, tb=tb, tb_offset=tb_offset
    )
    sys.stderr.write(template)


try:
    # third party
    from IPython import get_ipython

    get_ipython().set_custom_exc((SyftException,), syft_exception_handler)  # noqa: F821
except Exception:
    pass  # nosec


@serializable(canonical_name="SyftAttributeError", version=1)
class SyftAttributeError(AttributeError, SyftException):
    pass



# third party
import tomli
import tomli_w

# relative
from .rathole import RatholeConfig


class TomlReaderWriter:
    @staticmethod
    def load(toml_str: str) -> dict:
        return tomli.loads(toml_str)

    @staticmethod
    def dump(toml_dict: str) -> str:
        return tomli_w.dumps(toml_dict)


class RatholeBaseToml:
    filename: str

    def __init__(self, toml_str: str) -> None:
        self.toml_writer = TomlReaderWriter
        self.toml_str = toml_str

    def read(self) -> dict:
        return self.toml_writer.load(self.toml_str)

    def save(self, toml_dict: dict) -> None:
        self.toml_str = self.toml_writer.dump(toml_dict)

    def _validate(self) -> bool:
        raise NotImplementedError

    @property
    def is_valid(self) -> bool:
        return self._validate()


class RatholeClientToml(RatholeBaseToml):
    filename: str = "client.toml"

    def set_remote_addr(self, remote_host: str) -> None:
        """Add a new remote address to the client toml file."""

        toml = self.read()

        # Add the new remote address
        if "client" not in toml:
            toml["client"] = {}

        toml["client"]["remote_addr"] = remote_host

        self.save(toml)

    def clear_remote_addr(self) -> None:
        """Clear the remote address from the client toml file."""

        toml = self.read()

        # Clear the remote address
        if "client" not in toml:
            return

        toml["client"]["remote_addr"] = ""

        self.save(toml)

    def add_config(self, config: RatholeConfig) -> None:
        """Add a new config to the toml file."""

        toml = self.read()

        # Add the new config
        if "services" not in toml["client"]:
            toml["client"]["services"] = {}

        if config.uuid not in toml["client"]["services"]:
            toml["client"]["services"][config.uuid] = {}

        toml["client"]["services"][config.uuid] = {
            "token": config.secret_token,
            "local_addr": config.local_address,
        }

        self.save(toml)

    def remove_config(self, uuid: str) -> None:
        """Remove a config from the toml file."""

        toml = self.read()

        # Remove the config
        if "services" not in toml["client"]:
            return

        if uuid not in toml["client"]["services"]:
            return

        del toml["client"]["services"][uuid]

        self.save(toml)

    def update_config(self, config: RatholeConfig) -> None:
        """Update a config in the toml file."""

        toml = self.read()

        # Update the config
        if "services" not in toml["client"]:
            return

        if config.uuid not in toml["client"]["services"]:
            return

        toml["client"]["services"][config.uuid] = {
            "token": config.secret_token,
            "local_addr": config.local_address,
        }

        self.save(toml)

    def get_config(self, uuid: str) -> RatholeConfig | None:
        """Get a config from the toml file."""

        toml = self.read()

        # Get the config
        if "services" not in toml["client"]:
            return None

        if uuid not in toml["client"]["services"]:
            return None

        service = toml["client"]["services"][uuid]

        return RatholeConfig(
            uuid=uuid,
            secret_token=service["token"],
            local_addr_host=service["local_addr"].split(":")[0],
            local_addr_port=service["local_addr"].split(":")[1],
        )

    def _validate(self) -> bool:
        toml = self.read()

        if not toml["client"]["remote_addr"]:
            return False

        for uuid, config in toml["client"]["services"].items():
            if not uuid:
                return False

            if not config["token"] or not config["local_addr"]:
                return False

        return True


class RatholeServerToml(RatholeBaseToml):
    filename: str = "server.toml"

    def set_rathole_listener_addr(self, bind_addr: str) -> None:
        """Set the bind address in the server toml file."""

        toml = self.read()

        # Set the bind address
        toml["server"]["bind_addr"] = bind_addr

        self.save(toml)

    def get_rathole_listener_addr(self) -> str:
        """Get the bind address from the server toml file."""

        toml = self.read()

        return toml["server"]["bind_addr"]

    def add_config(self, config: RatholeConfig) -> None:
        """Add a new config to the toml file."""

        toml = self.read()

        # Add the new config
        if "services" not in toml["server"]:
            toml["server"]["services"] = {}

        if config.uuid not in toml["server"]["services"]:
            toml["server"]["services"][config.uuid] = {}

        toml["server"]["services"][config.uuid] = {
            "token": config.secret_token,
            "bind_addr": config.local_address,
        }

        self.save(toml)

    def remove_config(self, uuid: str) -> None:
        """Remove a config from the toml file."""

        toml = self.read()

        # Remove the config
        if "services" not in toml["server"]:
            return

        if uuid not in toml["server"]["services"]:
            return

        del toml["server"]["services"][uuid]

        self.save(toml)

    def update_config(self, config: RatholeConfig) -> None:
        """Update a config in the toml file."""

        toml = self.read()

        # Update the config
        if "services" not in toml["server"]:
            return

        if config.uuid not in toml["server"]["services"]:
            return

        toml["server"]["services"][config.uuid] = {
            "token": config.secret_token,
            "bind_addr": config.local_address,
        }

        self.save(toml)

    def _validate(self) -> bool:
        toml = self.read()

        if not toml["server"]["bind_addr"]:
            return False

        for uuid, config in toml["server"]["services"].items():
            if not uuid:
                return False

            if not config["token"] or not config["bind_addr"]:
                return False

        return True



# stdlib
import enum
import json

# relative
from ..assets import load_svg


class Icon(enum.Enum):
    SEARCH = "search.svg"
    CLIPBOARD = "clipboard.svg"
    TABLE = "table.svg"
    FOLDER = "folder.svg"
    REQUEST = "request.svg"
    ARROW = "arrow.svg"
    COPY = "copy.svg"
    INFO = "info.svg"

    @property
    def svg(self) -> str:
        return load_svg(self.value)

    @property
    def js_escaped_svg(self) -> str:
        """
        Required for in-line SVGs in JS
        TODO remove after refactoring table + templating
        """
        return json.dumps(self.svg)



# relative
from ...serde.serializable import serializable
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from ...types.blob_storage import BlobStorageEntry


@serializable(canonical_name="BlobStorageStash", version=1)
class BlobStorageStash(BaseUIDStoreStash):
    object_type = BlobStorageEntry
    settings: PartitionSettings = PartitionSettings(
        name=BlobStorageEntry.__canonical_name__, object_type=BlobStorageEntry
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)



# future
from __future__ import annotations

# stdlib
import logging
from typing import Any
from typing import ClassVar

# third party
from pydantic import field_validator
from pydantic import model_validator
from typing_extensions import Self

# relative
from ..client.client import SyftClient
from ..serde.serializable import serializable
from ..service.action.action_object import ActionObject
from ..service.action.action_object import TwinMode
from ..service.action.action_types import action_types
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..service.response import SyftWarning
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject
from .uid import UID

logger = logging.getLogger(__name__)


def to_action_object(obj: Any) -> ActionObject:
    if isinstance(obj, ActionObject):
        return obj

    if type(obj) in action_types:
        return action_types[type(obj)](syft_action_data_cache=obj)
    raise ValueError(f"{type(obj)} not in action_types")


@serializable()
class TwinObject(SyftObject):
    __canonical_name__ = "TwinObject"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_searchable__: ClassVar[list[str]] = []

    id: UID
    private_obj: ActionObject
    private_obj_id: UID = None  # type: ignore
    mock_obj: ActionObject
    mock_obj_id: UID = None  # type: ignore

    @field_validator("private_obj", mode="before")
    @classmethod
    def make_private_obj(cls, v: Any) -> ActionObject:
        return to_action_object(v)

    @model_validator(mode="after")
    def make_private_obj_id(self) -> Self:
        if self.private_obj_id is None:
            self.private_obj_id = self.private_obj.id  # type: ignore[unreachable]
        return self

    @field_validator("mock_obj", mode="before")
    @classmethod
    def make_mock_obj(cls, v: Any) -> ActionObject:
        return to_action_object(v)

    @model_validator(mode="after")
    def make_mock_obj_id(self) -> Self:
        if self.mock_obj_id is None:
            self.mock_obj_id = self.mock_obj.id  # type: ignore[unreachable]
        return self

    @property
    def private(self) -> ActionObject:
        twin_id = self.id
        private = self.private_obj
        private.syft_twin_type = TwinMode.PRIVATE
        private.id = twin_id
        return private

    @property
    def mock(self) -> ActionObject:
        twin_id = self.id
        mock = self.mock_obj
        mock.syft_twin_type = TwinMode.MOCK
        mock.id = twin_id
        return mock

    def _save_to_blob_storage(
        self, allow_empty: bool = False
    ) -> SyftError | SyftSuccess | SyftWarning:
        # Set server location and verify key
        self.private_obj._set_obj_location_(
            self.syft_server_location,
            self.syft_client_verify_key,
        )
        self.mock_obj._set_obj_location_(
            self.syft_server_location,
            self.syft_client_verify_key,
        )
        mock_store_res = self.mock_obj._save_to_blob_storage(allow_empty=allow_empty)
        if isinstance(mock_store_res, SyftError):
            return mock_store_res
        return self.private_obj._save_to_blob_storage(allow_empty=allow_empty)

    def send(self, client: SyftClient, add_storage_permission: bool = True) -> Any:
        self._set_obj_location_(client.id, client.verify_key)
        blob_store_result = self._save_to_blob_storage()
        if isinstance(blob_store_result, SyftWarning):
            logger.debug(blob_store_result.message)
        res = client.api.services.action.set(
            self,
            add_storage_permission=add_storage_permission,
        )
        return res



"""Python Level API to launch Syft services."""

# future
from __future__ import annotations

# stdlib
from collections.abc import Callable
from enum import Enum
import getpass
import inspect
import logging
import os
import sys
from typing import Any

# third party
from IPython.display import display

# relative
from .abstract_server import ServerSideType
from .abstract_server import ServerType
from .client.client import login_as_guest as sy_login_as_guest
from .protocol.data_protocol import stage_protocol_changes
from .server.datasite import Datasite
from .server.enclave import Enclave
from .server.gateway import Gateway
from .server.uvicorn import serve_server
from .service.response import SyftError
from .service.response import SyftInfo
from .util.util import get_random_available_port

logger = logging.getLogger(__name__)

DEFAULT_PORT = 8080
DEFAULT_URL = "http://localhost"

ClientAlias = Any  # we don't want to import Client in case it changes


def get_server_type(server_type: str | ServerType | None) -> ServerType | None:
    if server_type is None:
        server_type = os.environ.get("ORCHESTRA_SERVER_TYPE", ServerType.DATASITE)
    try:
        return ServerType(server_type)
    except ValueError:
        print(f"server_type: {server_type} is not a valid ServerType: {ServerType}")
    return None


def get_deployment_type(deployment_type: str | None) -> DeploymentType | None:
    if deployment_type is None:
        deployment_type = os.environ.get(
            "ORCHESTRA_DEPLOYMENT_TYPE", DeploymentType.PYTHON
        )

    try:
        return DeploymentType(deployment_type)
    except ValueError:
        print(
            f"deployment_type: {deployment_type} is not a valid DeploymentType: {DeploymentType}"
        )
    return None


# Can also be specified by the environment variable
# ORCHESTRA_DEPLOYMENT_TYPE
class DeploymentType(Enum):
    PYTHON = "python"
    REMOTE = "remote"


class ServerHandle:
    def __init__(
        self,
        server_type: ServerType,
        deployment_type: DeploymentType,
        server_side_type: ServerSideType,
        name: str,
        port: int | None = None,
        url: str | None = None,
        python_server: Any | None = None,
        shutdown: Callable | None = None,
    ) -> None:
        self.server_type = server_type
        self.name = name
        self.port = port
        self.url = url
        self.python_server = python_server
        self.shutdown = shutdown
        self.deployment_type = deployment_type
        self.server_side_type = server_side_type

    @property
    def client(self) -> Any:
        if self.port:
            return sy_login_as_guest(url=self.url, port=self.port)  # type: ignore
        elif self.deployment_type == DeploymentType.PYTHON:
            return self.python_server.get_guest_client(verbose=False)  # type: ignore
        else:
            raise NotImplementedError(
                f"client not implemented for the deployment type:{self.deployment_type}"
            )

    def login_as_guest(self, **kwargs: Any) -> ClientAlias:
        return self.client.login_as_guest(**kwargs)

    def login(
        self, email: str | None = None, password: str | None = None, **kwargs: Any
    ) -> ClientAlias:
        if not email:
            email = input("Email: ")

        if not password:
            password = getpass.getpass("Password: ")

        return self.client.login(email=email, password=password, **kwargs)

    def register(
        self,
        name: str,
        email: str | None = None,
        password: str | None = None,
        password_verify: str | None = None,
        institution: str | None = None,
        website: str | None = None,
    ) -> Any:
        if not email:
            email = input("Email: ")
        if not password:
            password = getpass.getpass("Password: ")
        if not password_verify:
            password_verify = getpass.getpass("Confirm Password: ")
        if password != password_verify:
            return SyftError(message="Passwords do not match")

        client = self.client
        return client.register(
            name=name,
            email=email,
            password=password,
            institution=institution,
            password_verify=password_verify,
            website=website,
        )

    def land(self) -> None:
        if self.deployment_type == DeploymentType.PYTHON:
            if self.shutdown:
                self.shutdown()
        else:
            print(
                f"Shutdown not implemented for the deployment type:{self.deployment_type}",
                file=sys.stderr,
            )


def deploy_to_python(
    server_type_enum: ServerType,
    deployment_type_enum: DeploymentType,
    port: int | str,
    name: str,
    host: str,
    reset: bool,
    tail: bool,
    dev_mode: bool,
    processes: int,
    local_db: bool,
    server_side_type: ServerSideType,
    enable_warnings: bool,
    n_consumers: int,
    thread_workers: bool,
    create_producer: bool = False,
    queue_port: int | None = None,
    association_request_auto_approval: bool = False,
    background_tasks: bool = False,
    debug: bool = False,
    migrate: bool = False,
) -> ServerHandle:
    worker_classes = {
        ServerType.DATASITE: Datasite,
        ServerType.GATEWAY: Gateway,
        ServerType.ENCLAVE: Enclave,
    }

    if dev_mode:
        logger.debug("Staging Protocol Changes...")
        stage_protocol_changes()

    kwargs = {
        "name": name,
        "host": host,
        "port": port,
        "reset": reset,
        "processes": processes,
        "dev_mode": dev_mode,
        "tail": tail,
        "server_type": server_type_enum,
        "server_side_type": server_side_type,
        "enable_warnings": enable_warnings,
        "queue_port": queue_port,
        "n_consumers": n_consumers,
        "create_producer": create_producer,
        "association_request_auto_approval": association_request_auto_approval,
        "background_tasks": background_tasks,
        "debug": debug,
        "migrate": migrate,
    }

    if port:
        kwargs["in_memory_workers"] = True
        if port == "auto":
            port = get_random_available_port()
        else:
            try:
                port = int(port)
            except ValueError:
                raise ValueError(
                    f"port must be either 'auto' or a valid int not: {port}"
                )
        kwargs["port"] = port

        sig = inspect.signature(serve_server)
        supported_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}

        start, stop = serve_server(**supported_kwargs)
        start()
        return ServerHandle(
            server_type=server_type_enum,
            deployment_type=deployment_type_enum,
            name=name,
            port=port,
            url="http://localhost",
            shutdown=stop,
            server_side_type=server_side_type,
        )
    else:
        kwargs["local_db"] = local_db
        kwargs["thread_workers"] = thread_workers
        if server_type_enum in worker_classes:
            worker_class = worker_classes[server_type_enum]
            sig = inspect.signature(worker_class.named)
            supported_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}
            if "server_type" in sig.parameters.keys() and "migrate" in sig.parameters:
                supported_kwargs["migrate"] = migrate
            worker = worker_class.named(**supported_kwargs)
        else:
            raise NotImplementedError(
                f"server_type: {server_type_enum} is not supported"
            )

        def stop() -> None:
            worker.stop()

        return ServerHandle(
            server_type=server_type_enum,
            deployment_type=deployment_type_enum,
            name=name,
            python_server=worker,
            server_side_type=server_side_type,
            shutdown=stop,
        )


def deploy_to_remote(
    server_type_enum: ServerType,
    deployment_type_enum: DeploymentType,
    name: str,
    server_side_type: ServerSideType,
    host: str | None = None,
    port: int | None = None,
    migrate: bool = False,
) -> ServerHandle:
    if migrate:
        raise ValueError("Cannot migrate via orchestra on remote server")

    # Preference order: Environment Variable > Argument > Default
    server_url = os.getenv("SERVER_URL") or host or DEFAULT_URL
    server_port = os.getenv("SERVER_PORT") or port or DEFAULT_PORT
    if server_port == "auto":
        raise ValueError("Cannot use auto port on remote server")

    return ServerHandle(
        server_type=server_type_enum,
        deployment_type=deployment_type_enum,
        name=name,
        server_side_type=server_side_type,
        url=server_url,
        port=int(server_port),
    )


class Orchestra:
    @staticmethod
    def launch(
        # server information and deployment
        name: str | None = None,
        server_type: str | ServerType | None = None,
        deploy_to: str | None = None,
        server_side_type: str | None = None,
        # worker related inputs
        port: int | str | None = None,
        processes: int = 1,  # temporary work around for jax in subprocess
        local_db: bool = False,
        dev_mode: bool = False,
        reset: bool = False,
        tail: bool = False,
        host: str | None = "0.0.0.0",  # nosec
        enable_warnings: bool = False,
        n_consumers: int = 0,
        thread_workers: bool = False,
        create_producer: bool = False,
        queue_port: int | None = None,
        association_request_auto_approval: bool = False,
        background_tasks: bool = False,
        debug: bool = False,
        migrate: bool = False,
    ) -> ServerHandle:
        if dev_mode is True:
            thread_workers = True
        os.environ["DEV_MODE"] = str(dev_mode)

        server_type_enum: ServerType | None = get_server_type(server_type=server_type)
        server_side_type_enum = (
            ServerSideType.HIGH_SIDE
            if server_side_type is None
            else ServerSideType(server_side_type)
        )

        deployment_type_enum: DeploymentType | None = get_deployment_type(
            deployment_type=deploy_to
        )

        if deployment_type_enum == DeploymentType.PYTHON:
            server_handle = deploy_to_python(
                server_type_enum=server_type_enum,
                deployment_type_enum=deployment_type_enum,
                port=port,
                name=name,
                host=host,
                reset=reset,
                tail=tail,
                dev_mode=dev_mode,
                processes=processes,
                local_db=local_db,
                server_side_type=server_side_type_enum,
                enable_warnings=enable_warnings,
                n_consumers=n_consumers,
                thread_workers=thread_workers,
                create_producer=create_producer,
                queue_port=queue_port,
                association_request_auto_approval=association_request_auto_approval,
                background_tasks=background_tasks,
                debug=debug,
                migrate=migrate,
            )
            display(
                SyftInfo(
                    message=f"You have launched a development server at http://{host}:{server_handle.port}."
                    + "It is intended only for local use."
                )
            )
            return server_handle
        elif deployment_type_enum == DeploymentType.REMOTE:
            return deploy_to_remote(
                server_type_enum=server_type_enum,
                deployment_type_enum=deployment_type_enum,
                name=name,
                host=host,
                port=port,
                server_side_type=server_side_type_enum,
                migrate=migrate,
            )
        raise NotImplementedError(
            f"deployment_type: {deployment_type_enum} is not supported"
        )



# stdlib
from collections.abc import Callable
import inspect
import time
from typing import Any
from typing import cast

# relative
from ...serde.serializable import serializable
from ...server.worker_settings import WorkerSettings
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..code.user_code import UserCode
from ..context import AuthedServiceContext
from ..log.log_service import LogService
from ..queue.queue_stash import ActionQueueItem
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from .job_stash import Job
from .job_stash import JobStash
from .job_stash import JobStatus


def wait_until(
    predicate: Callable[[], bool], timeout: int = 10
) -> SyftSuccess | SyftError:
    start = time.time()
    code_string = inspect.getsource(predicate).strip()
    while time.time() - start < timeout:
        if predicate():
            return SyftSuccess(message=f"Predicate {code_string} is True")
        time.sleep(1)
    return SyftError(message=f"Timeout reached for predicate {code_string}")


@instrument
@serializable(canonical_name="JobService", version=1)
class JobService(AbstractService):
    store: DocumentStore
    stash: JobStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = JobStash(store=store)

    @service_method(
        path="job.get",
        name="get",
        roles=GUEST_ROLE_LEVEL,
    )
    def get(self, context: AuthedServiceContext, uid: UID) -> Job | SyftError:
        res = self.stash.get_by_uid(context.credentials, uid=uid)
        if res.is_err():
            return SyftError(message=res.err())
        else:
            res = res.ok()
            return res

    @service_method(path="job.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> list[Job] | SyftError:
        res = self.stash.get_all(context.credentials)
        if res.is_err():
            return SyftError(message=res.err())
        else:
            res = res.ok()
            return res

    @service_method(
        path="job.get_by_user_code_id",
        name="get_by_user_code_id",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_by_user_code_id(
        self, context: AuthedServiceContext, user_code_id: UID
    ) -> list[Job] | SyftError:
        res = self.stash.get_by_user_code_id(context.credentials, user_code_id)
        if res.is_err():
            return SyftError(message=res.err())

        res = res.ok()
        return res

    @service_method(
        path="job.delete",
        name="delete",
        roles=ADMIN_ROLE_LEVEL,
    )
    def delete(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        res = self.stash.delete_by_uid(context.credentials, uid)
        if res.is_err():
            return SyftError(message=res.err())
        return SyftSuccess(message="Great Success!")

    @service_method(
        path="job.get_by_result_id",
        name="get_by_result_id",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_by_result_id(
        self, context: AuthedServiceContext, result_id: UID
    ) -> Job | None | SyftError:
        res = self.stash.get_by_result_id(context.credentials, result_id)
        if res.is_err():
            return SyftError(message=res.err())
        return res.ok()

    @service_method(
        path="job.restart",
        name="restart",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def restart(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        job_or_err = self.stash.get_by_uid(context.credentials, uid=uid)
        if job_or_err.is_err():
            return SyftError(message=job_or_err.err())
        if job_or_err.ok() is None:
            return SyftError(message="Job not found")

        job = job_or_err.ok()
        if job.parent_job_id is not None:
            return SyftError(
                message="Not possible to restart subjobs. Please restart the parent job."
            )
        if job.status == JobStatus.PROCESSING:
            return SyftError(
                message="Jobs in progress cannot be restarted. "
                "Please wait for completion or cancel the job via .cancel() to proceed."
            )

        job.status = JobStatus.CREATED
        self.update(context=context, job=job)

        task_uid = UID()
        worker_settings = WorkerSettings.from_server(context.server)
        worker_pool_ref = context.server.get_worker_pool_ref_by_name(
            context.credentials
        )
        if isinstance(worker_pool_ref, SyftError):
            return worker_pool_ref

        queue_item = ActionQueueItem(
            id=task_uid,
            server_uid=context.server.id,
            syft_client_verify_key=context.credentials,
            syft_server_location=context.server.id,
            job_id=job.id,
            worker_settings=worker_settings,
            args=[],
            kwargs={"action": job.action},
            worker_pool=worker_pool_ref,
        )

        context.server.queue_stash.set_placeholder(context.credentials, queue_item)
        context.server.job_stash.set(context.credentials, job)

        log_service = context.server.get_service("logservice")
        result = log_service.restart(context, job.log_id)
        if isinstance(result, SyftError):
            return result

        return SyftSuccess(message="Great Success!")

    @service_method(
        path="job.update",
        name="update",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def update(
        self, context: AuthedServiceContext, job: Job
    ) -> SyftSuccess | SyftError:
        res = self.stash.update(context.credentials, obj=job)
        if res.is_err():
            return SyftError(message=res.err())
        res = res.ok()
        return SyftSuccess(message="Great Success!")

    def _kill(self, context: AuthedServiceContext, job: Job) -> SyftSuccess | SyftError:
        # set job and subjobs status to TERMINATING
        # so that MonitorThread can kill them
        job.status = JobStatus.TERMINATING
        res = self.stash.update(context.credentials, obj=job)
        results = [res]

        # attempt to kill all subjobs
        subjobs_or_err = self.stash.get_by_parent_id(context.credentials, uid=job.id)
        if subjobs_or_err.is_ok() and subjobs_or_err.ok() is not None:
            subjobs = subjobs_or_err.ok()
            for subjob in subjobs:
                subjob.status = JobStatus.TERMINATING
                res = self.stash.update(context.credentials, obj=subjob)
                results.append(res)

        errors = [res.err() for res in results if res.is_err()]
        if errors:
            return SyftError(message=f"Failed to kill job: {errors}")

        # wait for job and subjobs to be killed by MonitorThread
        wait_until(lambda: job.fetched_status == JobStatus.INTERRUPTED)
        wait_until(
            lambda: all(
                subjob.fetched_status == JobStatus.INTERRUPTED for subjob in job.subjobs
            )
        )

        return SyftSuccess(message="Job killed successfully!")

    @service_method(
        path="job.kill",
        name="kill",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def kill(self, context: AuthedServiceContext, id: UID) -> SyftSuccess | SyftError:
        job_or_err = self.stash.get_by_uid(context.credentials, uid=id)
        if job_or_err.is_err():
            return SyftError(message=job_or_err.err())
        if job_or_err.ok() is None:
            return SyftError(message="Job not found")

        job = job_or_err.ok()
        if job.parent_job_id is not None:
            return SyftError(
                message="Not possible to cancel subjobs. To stop execution, please cancel the parent job."
            )
        if job.status != JobStatus.PROCESSING:
            return SyftError(message="Job is not running")
        if job.job_pid is None:
            return SyftError(
                message="Job termination disabled in dev mode. "
                "Set 'dev_mode=False' or 'thread_workers=False' to enable."
            )

        return self._kill(context, job)

    @service_method(
        path="job.get_subjobs",
        name="get_subjobs",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_subjobs(
        self, context: AuthedServiceContext, uid: UID
    ) -> list[Job] | SyftError:
        res = self.stash.get_by_parent_id(context.credentials, uid=uid)
        if res.is_err():
            return SyftError(message=res.err())
        else:
            return res.ok()

    @service_method(
        path="job.get_active", name="get_active", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_active(self, context: AuthedServiceContext) -> list[Job] | SyftError:
        res = self.stash.get_active(context.credentials)
        if res.is_err():
            return SyftError(message=res.err())
        return res.ok()

    @service_method(
        path="job.add_read_permission_job_for_code_owner",
        name="add_read_permission_job_for_code_owner",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def add_read_permission_job_for_code_owner(
        self, context: AuthedServiceContext, job: Job, user_code: UserCode
    ) -> None:
        permission = ActionObjectPermission(
            job.id, ActionPermission.READ, user_code.user_verify_key
        )
        return self.stash.add_permission(permission=permission)

    @service_method(
        path="job.add_read_permission_log_for_code_owner",
        name="add_read_permission_log_for_code_owner",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def add_read_permission_log_for_code_owner(
        self, context: AuthedServiceContext, log_id: UID, user_code: UserCode
    ) -> Any:
        log_service = context.server.get_service("logservice")
        log_service = cast(LogService, log_service)
        return log_service.stash.add_permission(
            ActionObjectPermission(
                log_id, ActionPermission.READ, user_code.user_verify_key
            )
        )

    @service_method(
        path="job.create_job_for_user_code_id",
        name="create_job_for_user_code_id",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def create_job_for_user_code_id(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        result: ActionObject | None = None,
        log_stdout: str = "",
        log_stderr: str = "",
        status: JobStatus = JobStatus.CREATED,
        add_code_owner_read_permissions: bool = True,
    ) -> Job | SyftError:
        is_resolved = status in [JobStatus.COMPLETED, JobStatus.ERRORED]
        job = Job(
            id=UID(),
            server_uid=context.server.id,
            action=None,
            result=result,
            status=status,
            parent_id=None,
            log_id=UID(),
            job_pid=None,
            user_code_id=user_code_id,
            resolved=is_resolved,
        )
        user_code_service = context.server.get_service("usercodeservice")
        user_code = user_code_service.get_by_uid(context=context, uid=user_code_id)
        if isinstance(user_code, SyftError):
            return user_code

        # The owner of the code should be able to read the job
        self.stash.set(context.credentials, job)

        log_service = context.server.get_service("logservice")
        res = log_service.add(
            context,
            job.log_id,
            job.id,
            stdout=log_stdout,
            stderr=log_stderr,
        )
        if isinstance(res, SyftError):
            return res

        if add_code_owner_read_permissions:
            self.add_read_permission_job_for_code_owner(context, job, user_code)
            self.add_read_permission_log_for_code_owner(context, job.log_id, user_code)

        return job


TYPE_TO_SERVICE[Job] = JobService



# stdlib
import re
from types import MethodType
from typing import Any

# relative
from ..types.dicttuple import DictTuple
from ..types.syft_object import SyftObject
from .util import sanitize_html


def _patch_ipython_sanitization() -> None:
    try:
        # third party
        from IPython import get_ipython
    except ImportError:
        return

    ip = get_ipython()
    if ip is None:
        return

    # stdlib
    from importlib import resources

    # relative
    from .assets import load_css
    from .assets import load_js
    from .notebook_ui.components.sync import ALERT_CSS
    from .notebook_ui.components.sync import COPY_CSS
    from .notebook_ui.styles import CSS_CODE
    from .notebook_ui.styles import FONT_CSS
    from .notebook_ui.styles import ITABLES_CSS
    from .notebook_ui.styles import JS_DOWNLOAD_FONTS

    tabulator_js = load_js("tabulator.min.js")
    tabulator_js = tabulator_js.replace(
        "define(t)", "define('tabulator-tables', [], t)"
    )

    SKIP_SANITIZE = [
        FONT_CSS,
        ITABLES_CSS,
        CSS_CODE,
        JS_DOWNLOAD_FONTS,
        tabulator_js,
        load_css("tabulator_pysyft.min.css"),
        load_js("table.js"),
    ]

    css_reinsert = f"""
<style>{FONT_CSS}</style>
<style>{ITABLES_CSS}</style>
{JS_DOWNLOAD_FONTS}
{CSS_CODE}
<style>{ALERT_CSS}</style>
<style>{COPY_CSS}</style>
"""

    escaped_js_css = re.compile(
        "|".join(re.escape(substr) for substr in SKIP_SANITIZE),
        re.IGNORECASE | re.MULTILINE,
    )

    table_template = (
        resources.files("syft.assets.jinja").joinpath("table.jinja2").read_text()
    )
    table_template = table_template.strip()
    table_template = re.sub(r"\\{\\{.*?\\}\\}", ".*?", re.escape(table_template))
    escaped_template = re.compile(table_template, re.DOTALL | re.VERBOSE)

    jobs_repr_template = (
        r"<!-- Start job_repr_template -->(.*?)<!-- End job_repr_template -->"
    )
    jobs_pattern = re.compile(jobs_repr_template, re.DOTALL)

    def display_sanitized_html(obj: SyftObject | DictTuple) -> str | None:
        if callable(getattr(obj, "_repr_html_", None)):
            html_str = obj._repr_html_()
            if html_str is not None:
                matching_table = escaped_template.findall(html_str)
                matching_jobs = jobs_pattern.findall(html_str)
                template = "\n".join(matching_table + matching_jobs)
                sanitized_str = escaped_template.sub("", html_str)
                sanitized_str = escaped_js_css.sub("", sanitized_str)
                sanitized_str = jobs_pattern.sub("", sanitized_str)
                sanitized_str = sanitize_html(sanitized_str)
                return f"{css_reinsert} {sanitized_str} {template}"
        return None

    def display_sanitized_md(obj: SyftObject) -> str | None:
        if callable(getattr(obj, "_repr_markdown_", None)):
            md = obj._repr_markdown_()
            if md is not None:
                return sanitize_html(md)
        return None

    ip.display_formatter.formatters["text/html"].for_type(
        SyftObject, display_sanitized_html
    )
    ip.display_formatter.formatters["text/html"].for_type(
        DictTuple, display_sanitized_html
    )
    ip.display_formatter.formatters["text/markdown"].for_type(
        SyftObject, display_sanitized_md
    )


def _patch_ipython_autocompletion() -> None:
    try:
        # third party
        from IPython import get_ipython
        from IPython.core.guarded_eval import EVALUATION_POLICIES
    except ImportError:
        return

    ipython = get_ipython()
    if ipython is None:
        return

    try:
        # this allows property getters to be used in nested autocomplete
        ipython.Completer.evaluation = "limited"
        ipython.Completer.use_jedi = False
        policy = EVALUATION_POLICIES["limited"]

        policy.allowed_getattr_external.update(
            [
                ("syft.client.api", "APIModule"),
                ("syft.client.api", "SyftAPI"),
            ]
        )
        original_can_get_attr = policy.can_get_attr

        def patched_can_get_attr(value: Any, attr: str) -> bool:
            attr_name = "__syft_allow_autocomplete__"
            # first check if exist to prevent side effects
            if hasattr(value, attr_name) and attr in getattr(value, attr_name, []):
                if attr in dir(value):
                    return True
                else:
                    return False
            else:
                return original_can_get_attr(value, attr)

        policy.can_get_attr = patched_can_get_attr
    except Exception:
        print("Failed to patch ipython autocompletion for syft property getters")

    try:
        # this constraints the completions for autocomplete.
        # if __syft_dir__ is defined we only autocomplete those properties
        original_attr_matches = ipython.Completer.attr_matches

        def patched_attr_matches(self, text: str) -> list[str]:  # type: ignore
            res = original_attr_matches(text)
            m2 = re.match(r"(.+)\.(\w*)$", self.line_buffer)
            if not m2:
                return res
            expr, _ = m2.group(1, 2)
            obj = self._evaluate_expr(expr)
            if isinstance(obj, SyftObject) and hasattr(obj, "__syft_dir__"):
                # here we filter all autocomplete results to only contain those
                # defined in __syft_dir__, however the original autocomplete prefixes
                # have the full path, while __syft_dir__ only defines the attr
                attrs = set(obj.__syft_dir__())
                new_res = []
                for r in res:
                    splitted = r.split(".")
                    if len(splitted) > 1:
                        attr_name = splitted[-1]
                        if attr_name in attrs:
                            new_res.append(r)
                return new_res
            else:
                return res

        ipython.Completer.attr_matches = MethodType(
            patched_attr_matches, ipython.Completer
        )
    except Exception:
        print("Failed to patch syft autocompletion for __syft_dir__")


def patch_ipython() -> None:
    _patch_ipython_sanitization()
    _patch_ipython_autocompletion()



# future
from __future__ import annotations

# stdlib
import threading

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...store.dict_document_store import DictStoreConfig
from ...store.document_store import BasePartitionSettings
from ...store.document_store import DocumentStore
from ...store.document_store import StoreConfig
from ...types.syft_object import SyftObject
from ...types.twin_object import TwinObject
from ...types.uid import LineageID
from ...types.uid import UID
from ..response import SyftSuccess
from .action_object import is_action_data_empty
from .action_permissions import ActionObjectEXECUTE
from .action_permissions import ActionObjectOWNER
from .action_permissions import ActionObjectPermission
from .action_permissions import ActionObjectREAD
from .action_permissions import ActionObjectWRITE
from .action_permissions import ActionPermission
from .action_permissions import StoragePermission

lock = threading.RLock()


class ActionStore:
    pass


@serializable(canonical_name="KeyValueActionStore", version=1)
class KeyValueActionStore(ActionStore):
    """Generic Key-Value Action store.

    Parameters:
        store_config: StoreConfig
            Backend specific configuration, including connection configuration, database name, or client class type.
        root_verify_key: Optional[SyftVerifyKey]
            Signature verification key, used for checking access permissions.
    """

    def __init__(
        self,
        server_uid: UID,
        store_config: StoreConfig,
        root_verify_key: SyftVerifyKey | None = None,
        document_store: DocumentStore | None = None,
    ) -> None:
        self.server_uid = server_uid
        self.store_config = store_config
        self.settings = BasePartitionSettings(name="Action")
        self.data = self.store_config.backing_store(
            "data", self.settings, self.store_config
        )
        self.permissions = self.store_config.backing_store(
            "permissions", self.settings, self.store_config, ddtype=set
        )
        self.storage_permissions = self.store_config.backing_store(
            "storage_permissions", self.settings, self.store_config, ddtype=set
        )

        if root_verify_key is None:
            root_verify_key = SyftSigningKey.generate().verify_key
        self.root_verify_key = root_verify_key

        self.__user_stash = None
        if document_store is not None:
            # relative
            from ...service.user.user_stash import UserStash

            self.__user_stash = UserStash(store=document_store)

    def get(
        self, uid: UID, credentials: SyftVerifyKey, has_permission: bool = False
    ) -> Result[SyftObject, str]:
        uid = uid.id  # We only need the UID from LineageID or UID

        # if you get something you need READ permission
        read_permission = ActionObjectREAD(uid=uid, credentials=credentials)
        if has_permission or self.has_permission(read_permission):
            try:
                if isinstance(uid, LineageID):
                    syft_object = self.data[uid.id]
                elif isinstance(uid, UID):
                    syft_object = self.data[uid]
                else:
                    raise Exception(f"Unrecognized UID type: {type(uid)}")
                return Ok(syft_object)
            except Exception as e:
                return Err(f"Could not find item with uid {uid}, {e}")
        return Err(f"Permission: {read_permission} denied")

    def get_mock(self, uid: UID) -> Result[SyftObject, str]:
        uid = uid.id  # We only need the UID from LineageID or UID

        try:
            syft_object = self.data[uid]
            if isinstance(syft_object, TwinObject) and not is_action_data_empty(
                syft_object.mock
            ):
                return Ok(syft_object.mock)
            return Err("No mock")
        except Exception as e:
            return Err(f"Could not find item with uid {uid}, {e}")

    def get_pointer(
        self,
        uid: UID,
        credentials: SyftVerifyKey,
        server_uid: UID,
    ) -> Result[SyftObject, str]:
        uid = uid.id  # We only need the UID from LineageID or UID

        try:
            if uid in self.data:
                obj = self.data[uid]
                read_permission = ActionObjectREAD(uid=uid, credentials=credentials)

                # if you have permission you can have private data
                if self.has_permission(read_permission):
                    if isinstance(obj, TwinObject):
                        return Ok(obj.private.syft_point_to(server_uid))
                    return Ok(obj.syft_point_to(server_uid))

                # if its a twin with a mock anyone can have this
                if isinstance(obj, TwinObject):
                    return Ok(obj.mock.syft_point_to(server_uid))

                # finally worst case you get ActionDataEmpty so you can still trace
                return Ok(obj.as_empty().syft_point_to(server_uid))

            return Err("Permission denied")
        except Exception as e:
            return Err(str(e))

    def exists(self, uid: UID) -> bool:
        uid = uid.id  # We only need the UID from LineageID or UID

        return uid in self.data

    def set(
        self,
        uid: UID,
        credentials: SyftVerifyKey,
        syft_object: SyftObject,
        has_result_read_permission: bool = False,
        add_storage_permission: bool = True,
    ) -> Result[SyftSuccess, Err]:
        uid = uid.id  # We only need the UID from LineageID or UID

        # if you set something you need WRITE permission
        write_permission = ActionObjectWRITE(uid=uid, credentials=credentials)
        can_write = self.has_permission(write_permission)

        if not self.exists(uid=uid):
            # attempt to claim it for writing
            if has_result_read_permission:
                ownership_result = self.take_ownership(uid=uid, credentials=credentials)
                can_write = True if ownership_result.is_ok() else False
            else:
                # root takes owneship, but you can still write
                ownership_result = self.take_ownership(
                    uid=uid, credentials=self.root_verify_key
                )
                can_write = True if ownership_result.is_ok() else False

        if can_write:
            self.data[uid] = syft_object
            if uid not in self.permissions:
                # create default permissions
                self.permissions[uid] = set()
            if has_result_read_permission:
                self.add_permission(ActionObjectREAD(uid=uid, credentials=credentials))
            else:
                self.add_permissions(
                    [
                        ActionObjectWRITE(uid=uid, credentials=credentials),
                        ActionObjectEXECUTE(uid=uid, credentials=credentials),
                    ]
                )

            if uid not in self.storage_permissions:
                # create default storage permissions
                self.storage_permissions[uid] = set()
            if add_storage_permission:
                self.add_storage_permission(
                    StoragePermission(uid=uid, server_uid=self.server_uid)
                )

            return Ok(SyftSuccess(message=f"Set for ID: {uid}"))
        return Err(f"Permission: {write_permission} denied")

    def take_ownership(
        self, uid: UID, credentials: SyftVerifyKey
    ) -> Result[SyftSuccess, str]:
        uid = uid.id  # We only need the UID from LineageID or UID

        # first person using this UID can claim ownership
        if uid not in self.permissions and uid not in self.data:
            self.add_permissions(
                [
                    ActionObjectOWNER(uid=uid, credentials=credentials),
                    ActionObjectWRITE(uid=uid, credentials=credentials),
                    ActionObjectREAD(uid=uid, credentials=credentials),
                    ActionObjectEXECUTE(uid=uid, credentials=credentials),
                ]
            )
            return Ok(SyftSuccess(message=f"Ownership of ID: {uid} taken."))
        return Err(f"UID: {uid} already owned.")

    def delete(self, uid: UID, credentials: SyftVerifyKey) -> Result[SyftSuccess, str]:
        uid = uid.id  # We only need the UID from LineageID or UID

        # if you delete something you need OWNER permission
        # is it bad to evict a key and have someone else reuse it?
        # perhaps we should keep permissions but no data?
        owner_permission = ActionObjectOWNER(uid=uid, credentials=credentials)
        if self.has_permission(owner_permission):
            if uid in self.data:
                del self.data[uid]
            if uid in self.permissions:
                del self.permissions[uid]
            return Ok(SyftSuccess(message=f"ID: {uid} deleted"))
        return Err(f"Permission: {owner_permission} denied")

    def has_permission(self, permission: ActionObjectPermission) -> bool:
        if not isinstance(permission.permission, ActionPermission):
            raise Exception(f"ObjectPermission type: {permission.permission} not valid")

        if (
            permission.credentials is not None
            and self.root_verify_key.verify == permission.credentials.verify
        ):
            return True

        if self.__user_stash is not None:
            # relative
            from ...service.user.user_roles import ServiceRole

            res = self.__user_stash.get_by_verify_key(
                credentials=permission.credentials,
                verify_key=permission.credentials,
            )

            if (
                res.is_ok()
                and (user := res.ok()) is not None
                and user.role in (ServiceRole.DATA_OWNER, ServiceRole.ADMIN)
            ):
                return True

        if (
            permission.uid in self.permissions
            and permission.permission_string in self.permissions[permission.uid]
        ):
            return True

        # 🟡 TODO 14: add ALL_READ, ALL_EXECUTE etc
        if permission.permission == ActionPermission.OWNER:
            pass
        elif permission.permission == ActionPermission.READ:
            pass
        elif permission.permission == ActionPermission.WRITE:
            pass
        elif permission.permission == ActionPermission.EXECUTE:
            pass

        return False

    def has_permissions(self, permissions: list[ActionObjectPermission]) -> bool:
        return all(self.has_permission(p) for p in permissions)

    def add_permission(self, permission: ActionObjectPermission) -> None:
        permissions = self.permissions[permission.uid]
        permissions.add(permission.permission_string)
        self.permissions[permission.uid] = permissions

    def remove_permission(self, permission: ActionObjectPermission) -> None:
        permissions = self.permissions[permission.uid]
        permissions.remove(permission.permission_string)
        self.permissions[permission.uid] = permissions

    def add_permissions(self, permissions: list[ActionObjectPermission]) -> None:
        for permission in permissions:
            self.add_permission(permission)

    def _get_permissions_for_uid(self, uid: UID) -> Result[set[str], str]:
        if uid in self.permissions:
            return Ok(self.permissions[uid])
        return Err(f"No permissions found for uid: {uid}")

    def get_all_permissions(self) -> Result[dict[UID, set[str]], str]:
        return Ok(dict(self.permissions.items()))

    def add_storage_permission(self, permission: StoragePermission) -> None:
        permissions = self.storage_permissions[permission.uid]
        permissions.add(permission.server_uid)
        self.storage_permissions[permission.uid] = permissions

    def add_storage_permissions(self, permissions: list[StoragePermission]) -> None:
        for permission in permissions:
            self.add_storage_permission(permission)

    def remove_storage_permission(self, permission: StoragePermission) -> None:
        permissions = self.storage_permissions[permission.uid]
        permissions.remove(permission.server_uid)
        self.storage_permissions[permission.uid] = permissions

    def has_storage_permission(self, permission: StoragePermission | UID) -> bool:
        if isinstance(permission, UID):
            permission = StoragePermission(uid=permission, server_uid=self.server_uid)

        if permission.uid in self.storage_permissions:
            return permission.server_uid in self.storage_permissions[permission.uid]
        return False

    def _get_storage_permissions_for_uid(self, uid: UID) -> Result[set[UID], str]:
        if uid in self.storage_permissions:
            return Ok(self.storage_permissions[uid])
        return Err(f"No storage permissions found for uid: {uid}")

    def get_all_storage_permissions(self) -> Result[dict[UID, set[UID]], str]:
        return Ok(dict(self.storage_permissions.items()))

    def _all(
        self,
        credentials: SyftVerifyKey,
        has_permission: bool | None = False,
    ) -> Result[list[SyftObject], str]:
        # this checks permissions
        res = [self.get(uid, credentials, has_permission) for uid in self.data.keys()]
        result = [x.ok() for x in res if x.is_ok()]
        return Ok(result)

    def migrate_data(
        self, to_klass: SyftObject, credentials: SyftVerifyKey
    ) -> Result[bool, str]:
        has_root_permission = credentials == self.root_verify_key

        if has_root_permission:
            for key, value in self.data.items():
                try:
                    if value.__canonical_name__ != to_klass.__canonical_name__:
                        continue
                    migrated_value = value.migrate_to(to_klass.__version__)
                except Exception as e:
                    return Err(
                        f"Failed to migrate data to {to_klass} {to_klass.__version__} for qk: {key}. Exception: {e}"
                    )
                result = self.set(
                    uid=key,
                    credentials=credentials,
                    syft_object=migrated_value,
                )

                if result.is_err():
                    return result.err()

            return Ok(True)

        return Err("You don't have permissions to migrate data.")


@serializable(canonical_name="DictActionStore", version=1)
class DictActionStore(KeyValueActionStore):
    """Dictionary-Based Key-Value Action store.

    Parameters:
        store_config: StoreConfig
            Backend specific configuration, including client class type.
        root_verify_key: Optional[SyftVerifyKey]
            Signature verification key, used for checking access permissions.
    """

    def __init__(
        self,
        server_uid: UID,
        store_config: StoreConfig | None = None,
        root_verify_key: SyftVerifyKey | None = None,
        document_store: DocumentStore | None = None,
    ) -> None:
        store_config = store_config if store_config is not None else DictStoreConfig()
        super().__init__(
            server_uid=server_uid,
            store_config=store_config,
            root_verify_key=root_verify_key,
            document_store=document_store,
        )


@serializable(canonical_name="SQLiteActionStore", version=1)
class SQLiteActionStore(KeyValueActionStore):
    """SQLite-Based Key-Value Action store.

    Parameters:
        store_config: StoreConfig
            SQLite specific configuration, including connection settings or client class type.
        root_verify_key: Optional[SyftVerifyKey]
            Signature verification key, used for checking access permissions.
    """

    pass


@serializable(canonical_name="MongoActionStore", version=1)
class MongoActionStore(KeyValueActionStore):
    """Mongo-Based  Action store.

    Parameters:
        store_config: StoreConfig
            Mongo specific configuration.
        root_verify_key: Optional[SyftVerifyKey]
            Signature verification key, used for checking access permissions.
    """

    pass



# stdlib
from collections.abc import Callable

# relative
from .syft_object import SyftMigrationRegistry
from .transforms import generate_transform_wrapper
from .transforms import validate_klass_and_version


def migrate(
    klass_from: type | str,
    klass_to: type | str,
    version_from: int | None = None,
    version_to: int | None = None,
) -> Callable:
    (
        klass_from_str,
        version_from,
        klass_to_str,
        version_to,
    ) = validate_klass_and_version(
        klass_from=klass_from,
        version_from=version_from,
        klass_to=klass_to,
        version_to=version_to,
    )

    if klass_from_str != klass_to_str:
        raise Exception(
            "Migration can only be performed across classes with same canonical name."
            f"Provided args: klass_from: {klass_from_str}, klass_to: {klass_to_str}"
        )

    if version_from is None or version_to is None:
        raise Exception(
            "Version information missing at either of the classes."
            f"{klass_from_str} has version: {version_from}, {klass_to_str} has version: {version_to}"
        )

    def decorator(function: Callable) -> Callable:
        transforms = function()

        wrapper = generate_transform_wrapper(
            klass_from=klass_from, klass_to=klass_to, transforms=transforms
        )

        SyftMigrationRegistry.register_migration_function(
            klass_type_str=klass_from_str,
            version_from=version_from,
            version_to=version_to,
            method=wrapper,
        )

        return function

    return decorator



# stdlib
from typing import Any

# relative
from ..abstract_server import ServerSideType
from ..abstract_server import ServerType
from ..serde.serializable import serializable
from ..server.credentials import SyftSigningKey
from ..service.metadata.server_metadata import ServerMetadataJSON
from ..service.network.server_peer import ServerPeer
from ..service.response import SyftError
from ..service.response import SyftException
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject
from ..util.assets import load_png_base64
from ..util.notebook_ui.styles import FONT_CSS
from .client import SyftClient
from .connection import ServerConnection


@serializable(canonical_name="GatewayClient", version=1)
class GatewayClient(SyftClient):
    # TODO: add widget repr for gateway client

    def proxy_to(self, peer: Any) -> SyftClient:
        # relative
        from .datasite_client import DatasiteClient
        from .enclave_client import EnclaveClient

        connection: type[ServerConnection] = self.connection.with_proxy(peer.id)
        metadata: ServerMetadataJSON | SyftError = connection.get_server_metadata(
            credentials=SyftSigningKey.generate()
        )
        if isinstance(metadata, SyftError):
            return metadata
        if metadata.server_type == ServerType.DATASITE.value:
            client_type: type[SyftClient] = DatasiteClient
        elif metadata.server_type == ServerType.ENCLAVE.value:
            client_type = EnclaveClient
        else:
            raise SyftException(
                f"Unknown server type {metadata.server_type} to create proxy client"
            )

        client = client_type(
            connection=connection,
            credentials=self.credentials,
        )
        return client

    def proxy_client_for(
        self,
        name: str,
        email: str | None = None,
        password: str | None = None,
        **kwargs: Any,
    ) -> SyftClient:
        peer = None
        if self.api.has_service("network"):
            peer = self.api.services.network.get_peer_by_name(name=name)
        if peer is None:
            return SyftError(message=f"No datasite with name {name}")
        res = self.proxy_to(peer)
        if email and password:
            res = res.login(email=email, password=password, **kwargs)
        return res

    @property
    def peers(self) -> list[ServerPeer] | SyftError | None:
        return ProxyClient(routing_client=self)

    @property
    def datasites(self) -> list[ServerPeer] | SyftError | None:
        return ProxyClient(routing_client=self, server_type=ServerType.DATASITE)

    @property
    def enclaves(self) -> list[ServerPeer] | SyftError | None:
        return ProxyClient(routing_client=self, server_type=ServerType.ENCLAVE)

    def _repr_html_(self) -> str:
        commands = """
        <li><span class='syft-code-block'>&lt;your_client&gt;
        .datasites</span> - list datasites connected to this gateway</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;
        .proxy_client_for</span> - get a connection to a listed datasite</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;
        .login</span> - log into the gateway</li>
        """

        command_list = f"""
        <ul style='padding-left: 1em;'>
            {commands}
        </ul>
        """

        small_server_symbol_logo = load_png_base64("small-syft-symbol-logo.png")

        url = getattr(self.connection, "url", None)
        server_details = f"<strong>URL:</strong> {url}<br />" if url else ""
        if self.metadata:
            server_details += f"<strong>Server Type:</strong> {self.metadata.server_type.capitalize()}<br />"
            server_side_type = (
                "Low Side"
                if self.metadata.server_side_type == ServerSideType.LOW_SIDE.value
                else "High Side"
            )
            server_details += (
                f"<strong>Server Side Type:</strong> {server_side_type}<br />"
            )
            server_details += (
                f"<strong>Syft Version:</strong> {self.metadata.syft_version}<br />"
            )

        return f"""
        <style>
            {FONT_CSS}

            .syft-container {{
                padding: 5px;
                font-family: 'Open Sans';
            }}
            .syft-alert-info {{
                color: #1F567A;
                background-color: #C2DEF0;
                border-radius: 4px;
                padding: 5px;
                padding: 13px 10px
            }}
            .syft-code-block {{
                background-color: #f7f7f7;
                border: 1px solid #cfcfcf;
                padding: 0px 2px;
            }}
            .syft-space {{
                margin-top: 1em;
            }}
        </style>
        <div class="syft-client syft-container">
            <img src="{small_server_symbol_logo}" alt="Logo"
            style="width:48px;height:48px;padding:3px;">
            <h2>Welcome to {self.name}</h2>
            <div class="syft-space">
                {server_details}
            </div>
            <div class='syft-alert-info syft-space'>
                &#9432;&nbsp;
                This server is run by the library PySyft to learn more about how it works visit
                <a href="https://github.com/OpenMined/PySyft">github.com/OpenMined/PySyft</a>.
            </div>
            <h4>Commands to Get Started</h4>
            {command_list}
        </div><br />
        """


class ProxyClient(SyftObject):
    __canonical_name__ = "ProxyClient"
    __version__ = SYFT_OBJECT_VERSION_1

    routing_client: GatewayClient
    server_type: ServerType | None = None

    def retrieve_servers(self) -> list[ServerPeer]:
        if self.server_type in [ServerType.DATASITE, ServerType.ENCLAVE]:
            return self.routing_client.api.services.network.get_peers_by_type(
                server_type=self.server_type
            )
        elif self.server_type is None:
            # if server type is None, return all servers
            return self.routing_client.api.services.network.get_all_peers()
        else:
            raise SyftException(
                f"Unknown server type {self.server_type} to retrieve proxy client"
            )

    def _repr_html_(self) -> str:
        return self.retrieve_servers()._repr_html_()

    def __len__(self) -> int:
        return len(self.retrieve_servers())

    def __getitem__(self, key: int | str) -> SyftClient:
        if not isinstance(key, int):
            raise SyftException(f"Key: {key} must be an integer")

        servers = self.retrieve_servers()

        if key >= len(servers):
            raise SyftException(f"Index {key} out of range for retrieved servers")

        return self.routing_client.proxy_to(servers[key])



# stdlib
from datetime import timedelta
from typing import Any
from typing import Optional

# third party
from pydantic import Field

# relative
from ...abstract_server import ServerSideType
from ...serde.serializable import serializable
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import LineageID
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.notebook_ui.components.sync import SyncTableObject
from ...util.notebook_ui.styles import FONT_CSS
from ...util.notebook_ui.styles import ITABLES_CSS
from ..context import AuthedServiceContext


def get_hierarchy_level_prefix(level: int) -> str:
    if level == 0:
        return ""
    else:
        return "--" * level + " "


class SyncStateRow(SyftObject):
    """A row in the SyncState table"""

    __canonical_name__ = "SyncStateItem"
    __version__ = SYFT_OBJECT_VERSION_1

    object: SyftObject
    previous_object: SyftObject | None = None
    current_state: str
    previous_state: str
    status: str
    level: int = 0
    last_sync_date: DateTime | None = None

    __syft_include_id_coll_repr__ = False

    # TODO table formatting
    __repr_attrs__ = [
        "previous_state",
        "current_state",
    ]
    __table_coll_widths__ = ["min-content", "auto", "auto", "auto"]

    def status_badge(self) -> dict[str, str]:
        status = self.status
        if status == "NEW":
            badge_color = "label-green"
        elif status == "SAME":
            badge_color = "label-gray"
        else:
            badge_color = "label-orange"
        return {"value": status.upper(), "type": badge_color}

    def _coll_repr_(self) -> dict[str, Any]:
        obj_view = SyncTableObject(object=self.object)

        if self.last_sync_date is not None:
            last_sync_date = self.last_sync_date
            last_sync_delta = timedelta(
                seconds=DateTime.now().utc_timestamp - last_sync_date.utc_timestamp
            )
            last_sync_delta_str = td_format(last_sync_delta)
            last_sync_html = (
                f"<p class='diff-state-no-obj'>{last_sync_delta_str} ago</p>"
            )
        else:
            last_sync_html = "<p class='diff-state-no-obj'>n/a</p>"
        return {
            "Status": self.status_badge(),
            "Summary": obj_view.to_html(),
            "Last Sync": last_sync_html,
        }

    @property
    def object_type(self) -> str:
        prefix = get_hierarchy_level_prefix(self.level)
        return f"{prefix}{type(self.object).__name__}"


def td_format(td_object: timedelta) -> str:
    seconds = int(td_object.total_seconds())
    if seconds == 0:
        return "0 seconds"

    periods = [
        ("year", 60 * 60 * 24 * 365),
        ("month", 60 * 60 * 24 * 30),
        ("day", 60 * 60 * 24),
        ("hour", 60 * 60),
        ("minute", 60),
        ("second", 1),
    ]

    strings = []
    for period_name, period_seconds in periods:
        if seconds >= period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            has_s = "s" if period_value > 1 else ""
            strings.append(f"{period_value} {period_name}{has_s}")

    return ", ".join(strings)


@serializable()
class SyncState(SyftObject):
    __canonical_name__ = "SyncState"
    __version__ = SYFT_OBJECT_VERSION_1

    server_uid: UID
    server_name: str
    server_side_type: ServerSideType
    objects: dict[UID, SyncableSyftObject] = {}
    dependencies: dict[UID, list[UID]] = {}
    created_at: DateTime = Field(default_factory=DateTime.now)
    previous_state_link: LinkedObject | None = None
    permissions: dict[UID, set[str]] = {}
    storage_permissions: dict[UID, set[UID]] = {}
    ignored_batches: dict[UID, int] = {}
    object_sync_dates: dict[UID, DateTime] = {}
    errors: dict[UID, str] = {}

    # NOTE importing ServerDiff annotation with TYPE_CHECKING does not work here,
    # since typing.get_type_hints does not check for TYPE_CHECKING-imported types
    _previous_state_diff: Any = None

    __attr_searchable__ = ["created_at"]

    def _set_previous_state_diff(self) -> None:
        # relative
        from .diff_state import ServerDiff

        # Re-use ServerDiff to compare to previous state
        # Low = previous state, high = current state
        # NOTE No previous sync state means everything is new
        previous_state = self.previous_state or SyncState(
            server_uid=self.server_uid,
            server_name=self.server_name,
            server_side_type=self.server_side_type,
            syft_client_verify_key=self.syft_client_verify_key,
        )
        self._previous_state_diff = ServerDiff.from_sync_state(
            previous_state, self, _include_server_status=False, direction=None
        )

    def get_previous_state_diff(self) -> Any:
        if self._previous_state_diff is None:
            self._set_previous_state_diff()

        return self._previous_state_diff

    @property
    def previous_state(self) -> Optional["SyncState"]:
        if self.previous_state_link is not None:
            return self.previous_state_link.resolve
        return None

    @property
    def all_ids(self) -> set[UID]:
        return set(self.objects.keys())

    def get_status(self, uid: UID) -> str | None:
        previous_state_diff = self.get_previous_state_diff()
        if previous_state_diff is None:
            return None
        diff = previous_state_diff.obj_uid_to_diff.get(uid)

        if diff is None:
            return None
        return diff.status

    def add_objects(
        self, objects: list[SyncableSyftObject], context: AuthedServiceContext
    ) -> None:
        for obj in objects:
            if isinstance(obj.id, LineageID):
                self.objects[obj.id.id] = obj
            elif isinstance(obj.id, UID):
                self.objects[obj.id] = obj
            else:
                raise ValueError(f"Unsupported id type: {type(obj.id)}")

        # TODO might get slow with large states,
        # need to build dependencies every time to not have UIDs
        # in dependencies that are not in objects
        self._build_dependencies(context=context)

    def _build_dependencies(self, context: AuthedServiceContext) -> None:
        self.dependencies = {}

        all_ids = self.all_ids
        for obj in self.objects.values():
            if hasattr(obj, "get_sync_dependencies"):
                deps = obj.get_sync_dependencies(context=context)
                deps = [d.id for d in deps if d.id in all_ids]  # type: ignore
                # TODO: Why is this en check here? here?
                if len(deps):
                    self.dependencies[obj.id.id] = deps

    @property
    def rows(self) -> list[SyncStateRow]:
        result = []
        ids = set()

        previous_diff = self.get_previous_state_diff()
        if previous_diff is None:
            raise ValueError("No previous state to compare to")
        for batch in previous_diff.batches:
            # NOTE we re-use ServerDiff to compare to previous state,
            # low_obj is previous state, high_obj is current state
            diff = batch.root_diff

            # If there is no high object, it means it was deleted
            # and we don't need to show it in the table
            if diff.high_obj is None:
                continue
            if diff.object_id in ids:
                continue

            ids.add(diff.object_id)
            row = SyncStateRow(
                object=diff.high_obj,
                previous_object=diff.low_obj,
                current_state=diff.diff_side_str("high"),
                previous_state=diff.diff_side_str("low"),
                level=0,  # TODO add levels to table
                status=batch.status,
                last_sync_date=diff.last_sync_date,
            )
            result.append(row)
        return result

    def _repr_html_(self) -> str:
        prop_template = (
            "<p class='paragraph'><strong><span class='pr-8'>{}: </span></strong>{}</p>"
        )
        name_html = prop_template.format("name", self.server_name)
        if self.previous_state_link is not None:
            previous_state = self.previous_state_link.resolve
            delta = timedelta(
                seconds=self.created_at.utc_timestamp
                - previous_state.created_at.utc_timestamp
            )
            val = f"{td_format(delta)} ago"
            date_html = prop_template.format("last sync", val)
        else:
            date_html = prop_template.format("last sync", "not synced yet")

        repr = f"""
        <style>
            {FONT_CSS}
            .syft-syncstate {{color: {SURFACE[options.color_theme]};}}
            .syft-syncstate h3,
            .syft-syncstate p
              {{font-family: 'Open Sans';}}
              {ITABLES_CSS}
              {{font-family: 'Open Sans';}}
              {ITABLES_CSS}
            </style>
        <div class='syft-syncstate'>
            <p style="margin-bottom:16px;"></p>
            {name_html}
            {date_html}
        </div>
"""
        return repr + self.rows._repr_html_()



# stdlib
import tempfile
from typing import Any

# relative
from .util import compatible_with_large_file_writes_capnp


def _serialize(
    obj: object,
    to_proto: bool = True,
    to_bytes: bool = False,
    for_hashing: bool = False,
) -> Any:
    # relative
    from .recursive import rs_object2proto

    proto = rs_object2proto(obj, for_hashing=for_hashing)
    if to_bytes:
        if compatible_with_large_file_writes_capnp(proto):
            with tempfile.TemporaryFile() as tmp_file:
                # Write data to a file to save RAM
                proto.write(tmp_file)
                # proto in memory, and bytes in file
                del proto
                # bytes in file
                tmp_file.seek(0)
                return tmp_file.read()
        else:
            res = proto.to_bytes()
            return res

    if to_proto:
        return proto



# future
from __future__ import annotations

# stdlib
from concurrent import futures
import json
import logging
import os
from typing import Any

# third party
import pandas as pd
import requests

# relative
from ..service.metadata.server_metadata import ServerMetadataJSON
from ..service.network.server_peer import ServerPeer
from ..service.network.server_peer import ServerPeerConnectionStatus
from ..service.response import SyftException
from ..types.server_url import ServerURL
from ..util.constants import DEFAULT_TIMEOUT
from .client import SyftClient as Client

logger = logging.getLogger(__name__)
NETWORK_REGISTRY_URL = (
    "https://raw.githubusercontent.com/OpenMined/NetworkRegistry/main/gateways.json"
)

NETWORK_REGISTRY_REPO = "https://github.com/OpenMined/NetworkRegistry"


def _get_all_networks(network_json: dict, version: str) -> list[dict]:
    return network_json.get(version, {}).get("gateways", [])


class NetworkRegistry:
    def __init__(self) -> None:
        self.all_networks: list[dict] = []

        try:
            network_json = self.load_network_registry_json()
            self.all_networks = _get_all_networks(
                network_json=network_json, version="2.0.0"
            )
        except Exception as e:
            logger.warning(
                f"Failed to get Network Registry, go checkout: {NETWORK_REGISTRY_REPO}. Exception: {e}"
            )

    @staticmethod
    def load_network_registry_json() -> dict:
        try:
            # Get the environment variable
            network_registry_json: str | None = os.getenv("NETWORK_REGISTRY_JSON")
            # If the environment variable exists, use it
            if network_registry_json is not None:
                network_json: dict = json.loads(network_registry_json)
            else:
                # Load the network registry from the NETWORK_REGISTRY_URL
                response = requests.get(NETWORK_REGISTRY_URL, timeout=30)  # nosec
                response.raise_for_status()  # raise an exception if the HTTP request returns an error
                network_json = response.json()

            return network_json

        except Exception as e:
            logger.warning(
                f"Failed to get Network Registry from {NETWORK_REGISTRY_REPO}. Exception: {e}"
            )
            return {}

    @property
    def online_networks(self) -> list[dict]:
        networks = self.all_networks

        def check_network(network: dict) -> dict[Any, Any] | None:
            url = "http://" + network["host_or_ip"] + ":" + str(network["port"]) + "/"
            try:
                res = requests.get(url, timeout=DEFAULT_TIMEOUT)  # nosec
                online = "This is a Syft Gateway server." in res.text
            except Exception:
                online = False

            # networks without frontend
            if not online:
                try:
                    ping_url = url + "api/v2/"
                    res = requests.get(ping_url, timeout=DEFAULT_TIMEOUT)  # nosec
                    online = res.status_code == 200
                except Exception:
                    online = False

            if online:
                version = network.get("version", None)
                # Check if syft version was described in NetworkRegistry
                # If it's unknown, try to update it to an available version.
                if not version or version == "unknown":
                    # If not defined, try to ask in /syft/version endpoint (supported by 0.7.0)
                    try:
                        version_url = url + "api/v2/metadata"
                        res = requests.get(version_url, timeout=DEFAULT_TIMEOUT)  # nosec
                        if res.status_code == 200:
                            network["version"] = res.json()["syft_version"]
                        else:
                            network["version"] = "unknown"
                    except Exception:
                        network["version"] = "unknown"
                return network
            return None

        # We can use a with statement to ensure threads are cleaned up promptly
        with futures.ThreadPoolExecutor(max_workers=20) as executor:
            # map
            _online_networks = list(
                executor.map(lambda network: check_network(network), networks)
            )

        return [network for network in _online_networks if network is not None]

    def _repr_html_(self) -> str:
        on = self.online_networks
        if len(on) == 0:
            return "(no gateways online - try syft.gateways.all_networks to see offline gateways)"
        df = pd.DataFrame(on)
        total_df = pd.DataFrame(
            [
                [
                    f"{len(on)} / {len(self.all_networks)} (online networks / all networks)"
                ]
                + [""] * (len(df.columns) - 1)
            ],
            columns=df.columns,
            index=["Total"],
        )
        df = pd.concat([df, total_df])
        return df._repr_html_()  # type: ignore

    def __repr__(self) -> str:
        on = self.online_networks
        if len(on) == 0:
            return "(no gateways online - try syft.gateways.all_networks to see offline gateways)"
        df = pd.DataFrame(on)
        total_df = pd.DataFrame(
            [
                [
                    f"{len(on)} / {len(self.all_networks)} (online networks / all networks)"
                ]
                + [""] * (len(df.columns) - 1)
            ],
            columns=df.columns,
            index=["Total"],
        )
        df = pd.concat([df, total_df])
        return df.to_string()

    def __len__(self) -> int:
        return len(self.all_networks)

    @staticmethod
    def create_client(network: dict[str, Any]) -> Client:
        # relative
        from ..client.client import connect

        try:
            port = int(network["port"])
            protocol = network["protocol"]
            host_or_ip = network["host_or_ip"]
            server_url = ServerURL(port=port, protocol=protocol, host_or_ip=host_or_ip)
            client = connect(url=str(server_url))
            return client.guest()
        except Exception as e:
            raise SyftException(f"Failed to login with: {network}. {e}")

    def __getitem__(self, key: str | int) -> Client:
        if isinstance(key, int):
            return self.create_client(network=self.online_networks[key])
        else:
            on = self.online_networks
            for network in on:
                if network["name"] == key:
                    return self.create_client(network=network)
        raise KeyError(f"Invalid key: {key} for {on}")


class DatasiteRegistry:
    def __init__(self) -> None:
        self.all_networks: list[dict] = []
        self.all_datasites: dict[str, ServerPeer] = {}
        try:
            network_json = NetworkRegistry.load_network_registry_json()
            self.all_networks = _get_all_networks(
                network_json=network_json, version="2.0.0"
            )
            self._get_all_datasites()
        except Exception as e:
            logger.warning(
                f"Failed to get Network Registry, go checkout: {NETWORK_REGISTRY_REPO}. {e}"
            )

    def _get_all_datasites(self) -> None:
        for network in self.all_networks:
            network_client = NetworkRegistry.create_client(network)
            datasites: list[ServerPeer] = network_client.datasites.retrieve_servers()
            for datasite in datasites:
                self.all_datasites[str(datasite.id)] = datasite

    @property
    def online_networks(self) -> list[dict]:
        networks = self.all_networks

        def check_network(network: dict) -> dict[Any, Any] | None:
            url = "http://" + network["host_or_ip"] + ":" + str(network["port"]) + "/"
            try:
                res = requests.get(url, timeout=DEFAULT_TIMEOUT)
                online = "This is a Syft Gateway server." in res.text
            except Exception:
                online = False

            # networks without frontend
            if not online:
                try:
                    ping_url = url + "api/v2/"
                    res = requests.get(ping_url, timeout=DEFAULT_TIMEOUT)
                    online = res.status_code == 200
                except Exception:
                    online = False

            if online:
                version = network.get("version", None)
                # Check if syft version was described in NetworkRegistry
                # If it's unknown, try to update it to an available version.
                if not version or version == "unknown":
                    # If not defined, try to ask in /syft/version endpoint (supported by 0.7.0)
                    try:
                        version_url = url + "api/v2/metadata"
                        res = requests.get(version_url, timeout=DEFAULT_TIMEOUT)
                        if res.status_code == 200:
                            network["version"] = res.json()["syft_version"]
                        else:
                            network["version"] = "unknown"
                    except Exception:
                        network["version"] = "unknown"
                return network

            return None

        # We can use a with statement to ensure threads are cleaned up promptly
        with futures.ThreadPoolExecutor(max_workers=20) as executor:
            # map
            _online_networks = list(
                executor.map(lambda network: check_network(network), networks)
            )

        return [network for network in _online_networks if network is not None]

    @property
    def online_datasites(self) -> list[tuple[ServerPeer, ServerMetadataJSON | None]]:
        networks = self.online_networks

        _all_online_datasites = []
        for network in networks:
            try:
                network_client = NetworkRegistry.create_client(network)
            except Exception as e:
                logger.error(f"Error in creating network client {e}")
                continue

            datasites: list[ServerPeer] = network_client.datasites.retrieve_servers()
            for datasite in datasites:
                self.all_datasites[str(datasite.id)] = datasite

            _all_online_datasites += [
                (datasite, datasite.guest_client.metadata)
                for datasite in datasites
                if datasite.ping_status == ServerPeerConnectionStatus.ACTIVE
            ]

        return [datasite for datasite in _all_online_datasites if datasite is not None]

    def __make_dict__(self) -> list[dict[str, Any]]:
        on = self.online_datasites
        datasites: list[dict[str, Any]] = []
        for datasite, metadata in on:
            datasite_dict: dict[str, Any] = {}
            datasite_dict["name"] = datasite.name
            if metadata is not None:
                datasite_dict["organization"] = metadata.organization
                datasite_dict["version"] = metadata.syft_version
            route = None
            if len(datasite.server_routes) > 0:
                route = datasite.pick_highest_priority_route()
            datasite_dict["host_or_ip"] = route.host_or_ip if route else "-"
            datasite_dict["protocol"] = route.protocol if route else "-"
            datasite_dict["port"] = route.port if route else "-"
            datasite_dict["id"] = datasite.id
            datasites.append(datasite_dict)

        return datasites

    def _repr_html_(self) -> str:
        on: list[dict[str, Any]] = self.__make_dict__()
        if len(on) == 0:
            return "(no datasites online - try syft.datasites.all_datasites to see offline datasites)"
        df = pd.DataFrame(on)
        total_df = pd.DataFrame(
            [
                [
                    f"{len(on)} / {len(self.all_datasites)} (online datasites / all datasites)"
                ]
                + [""] * (len(df.columns) - 1)
            ],
            columns=df.columns,
            index=["Total"],
        )
        df = pd.concat([df, total_df])
        return df._repr_html_()  # type: ignore

    def __repr__(self) -> str:
        on: list[dict[str, Any]] = self.__make_dict__()
        if len(on) == 0:
            return "(no datasites online - try syft.datasites.all_datasites to see offline datasites)"
        df = pd.DataFrame(on)
        total_df = pd.DataFrame(
            [
                [
                    f"{len(on)} / {len(self.all_datasites)} (online datasites / all datasites)"
                ]
                + [""] * (len(df.columns) - 1)
            ],
            columns=df.columns,
            index=["Total"],
        )
        df = pd.concat([df, total_df])
        return df._repr_html_()  # type: ignore

    def create_client(self, peer: ServerPeer) -> Client:
        try:
            return peer.guest_client
        except Exception as e:
            raise SyftException(f"Failed to login to: {peer}. {e}")

    def __getitem__(self, key: str | int) -> Client:
        if isinstance(key, int):
            return self.create_client(self.online_datasites[key][0])
        else:
            on = self.online_datasites
            count = 0
            for datasite, _ in on:
                if datasite.name == key:
                    return self.create_client(self.online_datasites[count][0])
                count += 1
        raise KeyError(f"Invalid key: {key} for {on}")


ENCLAVE_REGISTRY_URL = (
    "https://raw.githubusercontent.com/OpenMined/NetworkRegistry/main/enclaves.json"
)
ENCLAVE_REGISTRY_REPO = "https://github.com/OpenMined/NetworkRegistry"


class EnclaveRegistry:
    def __init__(self) -> None:
        self.all_enclaves: list[dict] = []
        try:
            response = requests.get(ENCLAVE_REGISTRY_URL)  # nosec
            enclaves_json = response.json()
            self.all_enclaves = enclaves_json["2.0.0"]["enclaves"]
        except Exception as e:
            logger.warning(
                f"Failed to get Enclave Registry, go checkout: {ENCLAVE_REGISTRY_REPO}. {e}"
            )

    @property
    def online_enclaves(self) -> list[dict]:
        enclaves = self.all_enclaves

        def check_enclave(enclave: dict) -> dict[Any, Any] | None:
            url = "http://" + enclave["host_or_ip"] + ":" + str(enclave["port"]) + "/"
            try:
                res = requests.get(url, timeout=DEFAULT_TIMEOUT)  # nosec
                online = "OpenMined Enclave Server Running" in res.text
            except Exception:
                online = False

            if online:
                version = enclave.get("version", None)
                # Check if syft version was described in EnclaveRegistry
                # If it's unknown, try to update it to an available version.
                if not version or version == "unknown":
                    # If not defined, try to ask in /syft/version endpoint (supported by 0.7.0)
                    try:
                        version_url = url + "api/v2/metadata"
                        res = requests.get(version_url, timeout=DEFAULT_TIMEOUT)  # nosec
                        if res.status_code == 200:
                            enclave["version"] = res.json()["syft_version"]
                        else:
                            enclave["version"] = "unknown"
                    except Exception:
                        enclave["version"] = "unknown"
                return enclave
            return None

        # We can use a with statement to ensure threads are cleaned up promptly
        with futures.ThreadPoolExecutor(max_workers=20) as executor:
            # map
            _online_enclaves = list(
                executor.map(lambda enclave: check_enclave(enclave), enclaves)
            )

        online_enclaves = [each for each in _online_enclaves if each is not None]
        return online_enclaves

    def _repr_html_(self) -> str:
        on = self.online_enclaves
        if len(on) == 0:
            return "(no enclaves online - try syft.enclaves.all_enclaves to see offline enclaves)"
        return pd.DataFrame(on)._repr_html_()  # type: ignore

    def __repr__(self) -> str:
        on = self.online_enclaves
        if len(on) == 0:
            return "(no enclaves online - try syft.enclaves.all_enclaves to see offline enclaves)"
        return pd.DataFrame(on).to_string()

    @staticmethod
    def create_client(enclave: dict[str, Any]) -> Client:
        # relative
        from ..client.client import connect

        try:
            port = int(enclave["port"])
            protocol = enclave["protocol"]
            host_or_ip = enclave["host_or_ip"]
            server_url = ServerURL(port=port, protocol=protocol, host_or_ip=host_or_ip)
            client = connect(url=str(server_url))
            return client.guest()
        except Exception as e:
            raise SyftException(f"Failed to login with: {enclave}. {e}")

    def __getitem__(self, key: str | int) -> Client:
        if isinstance(key, int):
            return self.create_client(enclave=self.online_enclaves[key])
        else:
            on = self.online_enclaves
            for enclave in on:
                if enclave["name"] == key:
                    return self.create_client(enclave=enclave)
        raise KeyError(f"Invalid key: {key} for {on}")



# relative
from ..assets import load_css

FONT_CSS = load_css("fonts.css")
ITABLES_CSS = load_css("itables.css")
STYLESHEET_URLS = [
    "https://fonts.googleapis.com/css2?family=Karla:ital,wght@0,200;0,300;0,400;0,500;0,600;0,700;0,800;1,200;1,300;1,400;1,500;1,600;1,700;1,800&family=Open+Sans:ital,wght@0,300..800;1,300..800&display=swap",
    "https://fonts.cdnfonts.com/css/dejavu-sans-mono",
]
STYLESHEET_JS_CALLS = "\n".join([f'addStyleSheet("{s}")' for s in STYLESHEET_URLS])

JS_DOWNLOAD_FONTS = f"""
<script>
function addStyleSheet(fileName) {{
  var head = document.head;
  var link = document.createElement("link");

  link.type = "text/css";
  link.rel = "stylesheet";
  link.href = fileName;

  head.appendChild(link);
}}

{STYLESHEET_JS_CALLS}
</script>
"""

CSS_CODE = f"""
<style>
  {load_css("style.css")}
</style>
"""
